"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import salt.modules.at as at
import salt.utils.path
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class AtTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for the salt.modules.at module
    """

    def setup_loader_modules(self):
        return {at: {}}

    atq_output = {
        "jobs": [
            {
                "date": "2014-12-11",
                "job": 101,
                "queue": "A",
                "tag": "",
                "time": "19:48:47",
                "user": "B",
            }
        ]
    }

    @classmethod
    def tearDownClass(cls):
        del cls.atq_output

    def test_atq_not_available(self):
        """
        Tests the at.atq not available for any type of os_family.
        """
        with patch("salt.modules.at._cmd", MagicMock(return_value=None)):
            with patch.dict(at.__grains__, {"os_family": "RedHat"}):
                self.assertEqual(at.atq(), "'at.atq' is not available.")

            with patch.dict(at.__grains__, {"os_family": ""}):
                self.assertEqual(at.atq(), "'at.atq' is not available.")

    def test_atq_no_jobs_available(self):
        """
        Tests the no jobs available for any type of os_family.
        """
        with patch("salt.modules.at._cmd", MagicMock(return_value="")):
            with patch.dict(at.__grains__, {"os_family": "RedHat"}):
                self.assertDictEqual(at.atq(), {"jobs": []})

            with patch.dict(at.__grains__, {"os_family": ""}):
                self.assertDictEqual(at.atq(), {"jobs": []})

    def test_atq_list(self):
        """
        Tests the list all queued and running jobs.
        """
        with patch("salt.modules.at._cmd") as salt_modules_at__cmd_mock:
            salt_modules_at__cmd_mock.return_value = (
                "101\tThu Dec 11             19:48:47 2014 A B"
            )
            with patch.dict(at.__grains__, {"os_family": "", "os": ""}):
                self.assertDictEqual(
                    at.atq(),
                    {
                        "jobs": [
                            {
                                "date": "2014-12-11",
                                "job": 101,
                                "queue": "A",
                                "tag": "",
                                "time": "19:48:00",
                                "user": "B",
                            }
                        ]
                    },
                )

            salt_modules_at__cmd_mock.return_value = (
                "101\t2014-12-11             19:48:47 A B"
            )
            with patch.dict(at.__grains__, {"os_family": "RedHat", "os": ""}):
                self.assertDictEqual(
                    at.atq(),
                    {
                        "jobs": [
                            {
                                "date": "2014-12-11",
                                "job": 101,
                                "queue": "A",
                                "tag": "",
                                "time": "19:48:47",
                                "user": "B",
                            }
                        ]
                    },
                )

            salt_modules_at__cmd_mock.return_value = (
                "SALT: Dec 11,             2014 19:48 A 101 B"
            )
            with patch.dict(at.__grains__, {"os_family": "", "os": "OpenBSD"}):
                self.assertDictEqual(
                    at.atq(),
                    {
                        "jobs": [
                            {
                                "date": "2014-12-11",
                                "job": "101",
                                "queue": "B",
                                "tag": "",
                                "time": "19:48:00",
                                "user": "A",
                            }
                        ]
                    },
                )

    def test_atrm(self):
        """
        Tests for remove jobs from the queue.
        """
        with patch("salt.modules.at.atq", MagicMock(return_value=self.atq_output)):
            with patch.object(salt.utils.path, "which", return_value=None):
                self.assertEqual(at.atrm(), "'at.atrm' is not available.")

            with patch.object(salt.utils.path, "which", return_value=True):
                self.assertDictEqual(at.atrm(), {"jobs": {"removed": [], "tag": None}})

            with patch.object(at, "_cmd", return_value=True):
                with patch.object(salt.utils.path, "which", return_value=True):
                    self.assertDictEqual(
                        at.atrm("all"), {"jobs": {"removed": ["101"], "tag": None}}
                    )

            with patch.object(at, "_cmd", return_value=True):
                with patch.object(salt.utils.path, "which", return_value=True):
                    self.assertDictEqual(
                        at.atrm(101), {"jobs": {"removed": ["101"], "tag": None}}
                    )

            with patch.object(at, "_cmd", return_value=None):
                self.assertEqual(at.atrm(101), "'at.atrm' is not available.")

    def test_jobcheck(self):
        """
        Tests for check the job from queue.
        """
        with patch("salt.modules.at.atq", MagicMock(return_value=self.atq_output)):
            self.assertDictEqual(at.jobcheck(), {"error": "You have given a condition"})

            self.assertDictEqual(
                at.jobcheck(runas="foo"),
                {"note": "No match jobs or time format error", "jobs": []},
            )

            self.assertDictEqual(
                at.jobcheck(
                    runas="B", tag="", hour=19, minute=48, day=11, month=12, Year=2014
                ),
                {
                    "jobs": [
                        {
                            "date": "2014-12-11",
                            "job": 101,
                            "queue": "A",
                            "tag": "",
                            "time": "19:48:47",
                            "user": "B",
                        }
                    ]
                },
            )

    def test_at(self):
        """
        Tests for add a job to the queue.
        """
        with patch("salt.modules.at.atq", MagicMock(return_value=self.atq_output)):
            self.assertDictEqual(at.at(), {"jobs": []})

            with patch.object(salt.utils.path, "which", return_value=None):
                self.assertEqual(
                    at.at("12:05am", "/sbin/reboot", tag="reboot"),
                    "'at.at' is not available.",
                )

            with patch.object(salt.utils.path, "which", return_value=True):
                with patch.dict(at.__grains__, {"os_family": "RedHat"}):
                    mock = MagicMock(return_value=None)
                    with patch.dict(at.__salt__, {"cmd.run": mock}):
                        self.assertEqual(
                            at.at("12:05am", "/sbin/reboot", tag="reboot"),
                            "'at.at' is not available.",
                        )

                    mock = MagicMock(return_value="Garbled time")
                    with patch.dict(at.__salt__, {"cmd.run": mock}):
                        self.assertDictEqual(
                            at.at("12:05am", "/sbin/reboot", tag="reboot"),
                            {"jobs": [], "error": "invalid timespec"},
                        )

                    mock = MagicMock(return_value="warning: commands\nA B")
                    with patch.dict(at.__salt__, {"cmd.run": mock}):
                        with patch.dict(at.__grains__, {"os": "OpenBSD"}):
                            self.assertDictEqual(
                                at.at("12:05am", "/sbin/reboot", tag="reboot"),
                                {
                                    "jobs": [
                                        {
                                            "date": "2014-12-11",
                                            "job": 101,
                                            "queue": "A",
                                            "tag": "",
                                            "time": "19:48:47",
                                            "user": "B",
                                        }
                                    ]
                                },
                            )

                with patch.dict(at.__grains__, {"os_family": ""}):
                    mock = MagicMock(return_value=None)
                    with patch.dict(at.__salt__, {"cmd.run": mock}):
                        self.assertEqual(
                            at.at("12:05am", "/sbin/reboot", tag="reboot"),
                            "'at.at' is not available.",
                        )

    def test_atc(self):
        """
        Tests for atc
        """
        with patch.object(at, "_cmd", return_value=None):
            self.assertEqual(at.atc(101), "'at.atc' is not available.")

        with patch.object(at, "_cmd", return_value=""):
            self.assertDictEqual(at.atc(101), {"error": "invalid job id '101'"})

        with patch.object(at, "_cmd", return_value="101\tThu Dec 11 19:48:47 2014 A B"):
            self.assertEqual(at.atc(101), "101\tThu Dec 11 19:48:47 2014 A B")



"""
Returners Directory

:func:`get_returner_options` is a general purpose function that returners may
use to fetch their configuration options.
"""

import logging

log = logging.getLogger(__name__)


def get_returner_options(virtualname=None, ret=None, attrs=None, **kwargs):
    """
    Get the returner options from salt.

    :param str virtualname: The returner virtualname (as returned
        by __virtual__()
    :param ret: result of the module that ran. dict-like object

        May contain a `ret_config` key pointing to a string
        If a `ret_config` is specified, config options are read from::

            value.virtualname.option

        If not, config options are read from::

            value.virtualname.option

    :param attrs: options the returner wants to read
    :param __opts__: Optional dict-like object that contains a fallback config
        in case the param `__salt__` is not supplied.

        Defaults to empty dict.
    :param __salt__: Optional dict-like object that exposes the salt API.

        Defaults to empty dict.

        a) if __salt__ contains a 'config.option' configuration options,
            we infer the returner is being called from a state or module run ->
            config is a copy of the `config.option` function

        b) if __salt__ was not available, we infer that the returner is being
        called from the Salt scheduler, so we look for the
        configuration options in the param `__opts__`
        -> cfg is a copy for the __opts__ dictionary
    :param str profile_attr: Optional.

        If supplied, an overriding config profile is read from
        the corresponding key of `__salt__`.

    :param dict profile_attrs: Optional

        .. fixme:: only keys are read

        For each key in profile_attr, a value is read in the are
        used to fetch a value pointed by 'virtualname.%key' in
        the dict found thanks to the param `profile_attr`
    """

    ret_config = _fetch_ret_config(ret)

    attrs = attrs or {}
    profile_attr = kwargs.get("profile_attr", None)
    profile_attrs = kwargs.get("profile_attrs", None)
    defaults = kwargs.get("defaults", None)
    __salt__ = kwargs.get("__salt__", {})
    __opts__ = kwargs.get("__opts__", {})

    # select the config source
    cfg = __salt__.get("config.option", __opts__)

    # browse the config for relevant options, store them in a dict
    _options = dict(
        _options_browser(
            cfg,
            ret_config,
            defaults,
            virtualname,
            attrs,
        )
    )

    # override some values with relevant profile options
    _options.update(
        _fetch_profile_opts(
            cfg, virtualname, __salt__, _options, profile_attr, profile_attrs
        )
    )

    # override some values with relevant options from
    # keyword arguments passed via return_kwargs
    if ret and "ret_kwargs" in ret:
        _options.update(ret["ret_kwargs"])

    return _options


def _fetch_ret_config(ret):
    """
    Fetches 'ret_config' if available.

    @see :func:`get_returner_options`
    """
    if not ret:
        return None
    if "ret_config" not in ret:
        return ""
    return str(ret["ret_config"])


def _fetch_option(cfg, ret_config, virtualname, attr_name):
    """
    Fetch a given option value from the config.

    @see :func:`get_returner_options`
    """
    # c_cfg is a dictionary returned from config.option for
    # any options configured for this returner.
    if isinstance(cfg, dict):
        c_cfg = cfg
    else:
        c_cfg = cfg("{}".format(virtualname), {})

    default_cfg_key = "{}.{}".format(virtualname, attr_name)
    if not ret_config:
        # Using the default configuration key
        if isinstance(cfg, dict):
            if default_cfg_key in cfg:
                return cfg[default_cfg_key]
            else:
                return c_cfg.get(attr_name)
        else:
            return c_cfg.get(attr_name, cfg(default_cfg_key))

    # Using ret_config to override the default configuration key
    ret_cfg = cfg("{}.{}".format(ret_config, virtualname), {})

    override_default_cfg_key = "{}.{}.{}".format(
        ret_config,
        virtualname,
        attr_name,
    )
    override_cfg_default = cfg(override_default_cfg_key)

    # Look for the configuration item in the override location
    ret_override_cfg = ret_cfg.get(attr_name, override_cfg_default)
    if ret_override_cfg:
        return ret_override_cfg

    # if not configuration item found, fall back to the default location.
    return c_cfg.get(attr_name, cfg(default_cfg_key))


def _options_browser(cfg, ret_config, defaults, virtualname, options):
    """
    Iterator generating all duples ```option name -> value```

    @see :func:`get_returner_options`
    """

    for option in options:

        # default place for the option in the config
        value = _fetch_option(cfg, ret_config, virtualname, options[option])

        if value:
            yield option, value
            continue

        # Attribute not found, check for a default value
        if defaults:
            if option in defaults:
                log.debug("Using default for %s %s", virtualname, option)
                yield option, defaults[option]
                continue

        # fallback (implicit else for all ifs)
        continue


def _fetch_profile_opts(
    cfg, virtualname, __salt__, _options, profile_attr, profile_attrs
):
    """
    Fetches profile specific options if applicable

    @see :func:`get_returner_options`

    :return: a options dict
    """

    if (not profile_attr) or (profile_attr not in _options):
        return {}

    # Using a profile and it is in _options

    creds = {}
    profile = _options[profile_attr]
    if profile:
        log.debug("Using profile %s", profile)

        if "config.option" in __salt__:
            creds = cfg(profile)
        else:
            creds = cfg.get(profile)

    if not creds:
        return {}

    return {
        pattr: creds.get("{}.{}".format(virtualname, profile_attrs[pattr]))
        for pattr in profile_attrs
    }



import ctypes
import logging
import multiprocessing

import pytest
import salt.config
import salt.exceptions
import salt.ext.tornado.gen
import salt.log.setup
import salt.master
import salt.transport.client
import salt.transport.server
import salt.utils.platform
import salt.utils.process
import salt.utils.stringutils
from saltfactories.utils.processes import terminate_process

log = logging.getLogger(__name__)


class ReqServerChannelProcess(salt.utils.process.SignalHandlingProcess):
    def __init__(self, config, req_channel_crypt):
        super().__init__()
        self._closing = False
        self.config = config
        self.req_channel_crypt = req_channel_crypt
        self.process_manager = salt.utils.process.ProcessManager(
            name="ReqServer-ProcessManager"
        )
        self.req_server_channel = salt.transport.server.ReqServerChannel.factory(
            self.config
        )
        self.req_server_channel.pre_fork(self.process_manager)
        self.io_loop = None
        self.running = multiprocessing.Event()

    def run(self):
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(
                    salt.crypt.Crypticle.generate_key_string()
                ),
            ),
            "serial": multiprocessing.Value(
                ctypes.c_longlong, lock=False  # We'll use the lock from 'secret'
            ),
        }

        self.io_loop = salt.ext.tornado.ioloop.IOLoop()
        self.io_loop.make_current()
        self.req_server_channel.post_fork(self._handle_payload, io_loop=self.io_loop)
        self.io_loop.add_callback(self.running.set)
        try:
            self.io_loop.start()
        except KeyboardInterrupt:
            pass

    def _handle_signals(self, signum, sigframe):
        self.close()
        super()._handle_signals(signum, sigframe)

    def __enter__(self):
        self.start()
        self.running.wait()
        return self

    def __exit__(self, *args):
        self.close()
        self.terminate()

    def close(self):
        if self._closing:
            return
        self._closing = True
        if self.req_server_channel is not None:
            self.req_server_channel.close()
            self.req_server_channel = None
        if self.process_manager is not None:
            self.process_manager.terminate()
            # Really terminate any process still left behind
            for pid in self.process_manager._process_map:
                terminate_process(pid=pid, kill_children=True, slow_stop=False)
            self.process_manager = None

    @salt.ext.tornado.gen.coroutine
    def _handle_payload(self, payload):
        if self.req_channel_crypt == "clear":
            raise salt.ext.tornado.gen.Return((payload, {"fun": "send_clear"}))
        raise salt.ext.tornado.gen.Return((payload, {"fun": "send"}))


@pytest.fixture
def req_server_channel(salt_master, req_channel_crypt):
    req_server_channel_process = ReqServerChannelProcess(
        salt_master.config.copy(), req_channel_crypt
    )
    try:
        with req_server_channel_process:
            yield
    finally:
        terminate_process(
            pid=req_server_channel_process.pid, kill_children=True, slow_stop=False
        )


def req_channel_crypt_ids(value):
    return "ReqChannel(crypt='{}')".format(value)


@pytest.fixture(params=["clear", "aes"], ids=req_channel_crypt_ids)
def req_channel_crypt(request):
    return request.param


@pytest.fixture
def req_channel(req_server_channel, salt_minion, req_channel_crypt):
    with salt.transport.client.ReqChannel.factory(
        salt_minion.config, crypt=req_channel_crypt
    ) as _req_channel:
        try:
            yield _req_channel
        finally:
            # Force termination of singleton
            _req_channel.obj._refcount = 0


def test_basic(req_channel):
    """
    Test a variety of messages, make sure we get the expected responses
    """
    msgs = [
        {"foo": "bar"},
        {"bar": "baz"},
        {"baz": "qux", "list": [1, 2, 3]},
    ]
    for msg in msgs:
        ret = req_channel.send(dict(msg), timeout=5, tries=1)
        assert ret["load"] == msg


def test_normalization(req_channel):
    """
    Since we use msgpack, we need to test that list types are converted to lists
    """
    types = {
        "list": list,
    }
    msgs = [
        {"list": tuple([1, 2, 3])},
    ]
    for msg in msgs:
        ret = req_channel.send(msg, timeout=5, tries=1)
        for key, value in ret["load"].items():
            assert types[key] == type(value)


def test_badload(req_channel, req_channel_crypt):
    """
    Test a variety of bad requests, make sure that we get some sort of error
    """
    msgs = ["", [], tuple()]
    if req_channel_crypt == "clear":
        for msg in msgs:
            ret = req_channel.send(msg, timeout=5, tries=1)
            assert ret == "payload and load must be a dict"
    else:
        for msg in msgs:
            with pytest.raises(salt.exceptions.AuthenticationError):
                req_channel.send(msg, timeout=5, tries=1)



"""test for pillar csvpillar.py"""


import salt.pillar.csvpillar as csvpillar
from tests.support.mock import mock_open, patch
from tests.support.unit import TestCase


class CSVPillarTestCase(TestCase):
    def test_001_load_utf8_csv(self):
        fake_csv = "id,foo,bar\r\nminion1,foo1,bar1"
        fake_dict = {"id": "minion1", "foo": "foo1", "bar": "bar1"}
        fopen_mock = mock_open(fake_csv)
        with patch("salt.utils.files.fopen", fopen_mock):
            result = csvpillar.ext_pillar(
                mid="minion1",
                pillar=None,
                path="/fake/path/file.csv",
                idkey="id",
                namespace=None,
            )
            self.assertDictEqual(fake_dict, result)

    def test_002_load_utf8_csv_namespc(self):
        fake_csv = "id,foo,bar\r\nminion1,foo1,bar1"
        fake_dict = {"baz": {"id": "minion1", "foo": "foo1", "bar": "bar1"}}
        fopen_mock = mock_open(fake_csv)
        with patch("salt.utils.files.fopen", fopen_mock):
            result = csvpillar.ext_pillar(
                mid="minion1",
                pillar=None,
                path="/fake/path/file.csv",
                idkey="id",
                namespace="baz",
            )
            self.assertDictEqual(fake_dict, result)



"""
Test case for the YAML SDB module
"""


import salt.sdb.yaml as sdb
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class TestYamlRenderer(TestCase):
    """
    Test case for the YAML SDB module
    """

    def test_plaintext(self):
        """
        Retrieve a value from the top level of the dictionary
        """
        plain = {"foo": "bar"}
        with patch("salt.sdb.yaml._get_values", MagicMock(return_value=plain)):
            self.assertEqual(sdb.get("foo"), "bar")

    def test_nested(self):
        """
        Retrieve a value from a nested level of the dictionary
        """
        plain = {"foo": {"bar": "baz"}}
        with patch("salt.sdb.yaml._get_values", MagicMock(return_value=plain)):
            self.assertEqual(sdb.get("foo:bar"), "baz")

    def test_encrypted(self):
        """
        Assume the content is plaintext if GPG is not configured
        """
        plain = {"foo": "bar"}
        with patch("salt.sdb.yaml._decrypt", MagicMock(return_value=plain)):
            with patch("salt.sdb.yaml._get_values", MagicMock(return_value=None)):
                self.assertEqual(sdb.get("foo", profile={"gpg": True}), "bar")



import salt.fileserver.hgfs as hgfs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase, skipIf


class HgfsFileTest(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {hgfs: {}}

    def test_env_is_exposed(self):
        """
        test _env_is_exposed method when
        base is in whitelist
        """
        with patch.dict(
            hgfs.__opts__,
            {"hgfs_saltenv_whitelist": "base", "hgfs_saltenv_blacklist": ""},
        ):
            assert hgfs._env_is_exposed("base")

    def test_env_is_exposed_blacklist(self):
        """
        test _env_is_exposed method when
        base is in blacklist
        """
        with patch.dict(
            hgfs.__opts__,
            {"hgfs_saltenv_whitelist": "", "hgfs_saltenv_blacklist": "base"},
        ):
            assert not hgfs._env_is_exposed("base")

    @skipIf(not hgfs.HAS_HG, "hglib needs to be installed")
    def test_fix_58852(self):
        """
        test to make sure python 3 can init hgfs
        """
        with patch.dict(
            hgfs.__opts__,
            {
                "cachedir": "/tmp/barf",
                "hgfs_base": "fnord",
                "hgfs_branch_method": "fnord",
                "hgfs_mountpoint": "fnord",
                "hgfs_root": "fnord",
                "hgfs_remotes": "fnord",
            },
        ):
            self.assertIsInstance(hgfs.init(), list)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.extfs as extfs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class ExtfsTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.extfs
    """

    def setup_loader_modules(self):
        return {extfs: {}}

    # 'mkfs' function tests: 1

    def test_mkfs(self):
        """
        Tests if a file system created on the specified device
        """
        mock = MagicMock()
        with patch.dict(extfs.__salt__, {"cmd.run": mock}):
            self.assertListEqual([], extfs.mkfs("/dev/sda1", "ext4"))

    # 'tune' function tests: 1

    def test_tune(self):
        """
        Tests if specified group was added
        """
        mock = MagicMock()
        with patch.dict(extfs.__salt__, {"cmd.run": mock}), patch(
            "salt.modules.extfs.tune", MagicMock(return_value="")
        ):
            self.assertEqual("", extfs.tune("/dev/sda1"))

    # 'dump' function tests: 1

    def test_dump(self):
        """
        Tests if specified group was added
        """
        mock = MagicMock()
        with patch.dict(extfs.__salt__, {"cmd.run": mock}):
            self.assertEqual({"attributes": {}, "blocks": {}}, extfs.dump("/dev/sda1"))

    # 'attributes' function tests: 1

    def test_attributes(self):
        """
        Tests if specified group was added
        """
        with patch(
            "salt.modules.extfs.dump",
            MagicMock(return_value={"attributes": {}, "blocks": {}}),
        ):
            self.assertEqual({}, extfs.attributes("/dev/sda1"))

    # 'blocks' function tests: 1

    def test_blocks(self):
        """
        Tests if specified group was added
        """
        with patch(
            "salt.modules.extfs.dump",
            MagicMock(return_value={"attributes": {}, "blocks": {}}),
        ):
            self.assertEqual({}, extfs.blocks("/dev/sda1"))



"""
Tests for the salt-run command
"""

import logging

import pytest
from tests.support.case import ShellCase
from tests.support.unit import skipIf

try:
    import libnacl.secret  # pylint: disable=unused-import
    import libnacl.sealed  # pylint: disable=unused-import

    HAS_LIBNACL = True
except (ImportError, OSError, AttributeError):
    HAS_LIBNACL = False

log = logging.getLogger(__name__)


@skipIf(not HAS_LIBNACL, "skipping test_nacl, libnacl is unavailable")
@pytest.mark.windows_whitelisted
class NaclTest(ShellCase):
    """
    Test the nacl runner
    """

    @pytest.mark.slow_test
    def test_keygen(self):
        """
        Test keygen
        """
        # Store the data
        ret = self.run_run_plus(
            "nacl.keygen",
        )
        self.assertIn("pk", ret["return"])
        self.assertIn("sk", ret["return"])

    @pytest.mark.slow_test
    def test_enc(self):
        """
        Test keygen
        """
        # Store the data
        ret = self.run_run_plus(
            "nacl.keygen",
        )
        self.assertIn("pk", ret["return"])
        self.assertIn("sk", ret["return"])
        pk = ret["return"]["pk"]
        sk = ret["return"]["sk"]

        unencrypted_data = "hello"

        # Encrypt with pk
        ret = self.run_run_plus(
            "nacl.enc",
            data=unencrypted_data,
            pk=pk,
        )
        self.assertIn("return", ret)

    @pytest.mark.slow_test
    def test_enc_dec(self):
        """
        Store, list, fetch, then flush data
        """
        # Store the data
        ret = self.run_run_plus(
            "nacl.keygen",
        )
        self.assertIn("pk", ret["return"])
        self.assertIn("sk", ret["return"])
        pk = ret["return"]["pk"]
        sk = ret["return"]["sk"]

        unencrypted_data = b"hello"

        # Encrypt with pk
        ret = self.run_run_plus(
            "nacl.enc",
            data=unencrypted_data,
            pk=pk,
        )
        self.assertIn("return", ret)
        encrypted_data = ret["return"]

        # Decrypt with sk
        ret = self.run_run_plus(
            "nacl.dec",
            data=encrypted_data,
            sk=sk,
        )
        self.assertIn("return", ret)
        self.assertEqual(unencrypted_data, ret["return"])

    @pytest.mark.slow_test
    def test_sealedbox_enc_dec(self):
        """
        Generate keys, encrypt, then decrypt.
        """
        # Store the data
        ret = self.run_run_plus(
            "nacl.keygen",
        )
        self.assertIn("pk", ret["return"])
        self.assertIn("sk", ret["return"])
        pk = ret["return"]["pk"]
        sk = ret["return"]["sk"]

        unencrypted_data = b"hello"

        # Encrypt with pk
        ret = self.run_run_plus(
            "nacl.sealedbox_encrypt",
            data=unencrypted_data,
            pk=pk,
        )
        encrypted_data = ret["return"]

        # Decrypt with sk
        ret = self.run_run_plus(
            "nacl.sealedbox_decrypt",
            data=encrypted_data,
            sk=sk,
        )
        self.assertEqual(unencrypted_data, ret["return"])

    @pytest.mark.slow_test
    def test_secretbox_enc_dec(self):
        """
        Generate keys, encrypt, then decrypt.
        """
        # Store the data
        ret = self.run_run_plus(
            "nacl.keygen",
        )
        self.assertIn("pk", ret["return"])
        self.assertIn("sk", ret["return"])
        pk = ret["return"]["pk"]
        sk = ret["return"]["sk"]

        unencrypted_data = b"hello"

        # Encrypt with pk
        ret = self.run_run_plus(
            "nacl.secretbox_encrypt",
            data=unencrypted_data,
            sk=sk,
        )
        encrypted_data = ret["return"]

        # Decrypt with sk
        ret = self.run_run_plus(
            "nacl.secretbox_decrypt",
            data=encrypted_data,
            sk=sk,
        )
        self.assertEqual(unencrypted_data, ret["return"])

    @pytest.mark.slow_test
    def test_enc_dec_no_pk_no_sk(self):
        """
        Store, list, fetch, then flush data
        """
        # Store the data
        ret = self.run_run_plus(
            "nacl.keygen",
        )
        self.assertIn("pk", ret["return"])
        self.assertIn("sk", ret["return"])
        pk = ret["return"]["pk"]
        sk = ret["return"]["sk"]

        unencrypted_data = b"hello"

        # Encrypt with pk
        ret = self.run_run_plus(
            "nacl.enc",
            data=unencrypted_data,
            pk=None,
        )
        self.assertIn("Exception: no pubkey or pk_file found", ret["return"])

        self.assertIn("return", ret)
        encrypted_data = ret["return"]

        # Decrypt with sk
        ret = self.run_run_plus(
            "nacl.dec",
            data=encrypted_data,
            sk=None,
        )
        self.assertIn("Exception: no key or sk_file found", ret["return"])



"""
Tests for the file state
"""
import pytest
import salt.states.file


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_get_source_sum_verify_ssl_false(
    salt_call_cli, tmpdir, ssl_webserver, verify_ssl
):
    """
    test verify_ssl with get_source_sum
    """
    web_file = ssl_webserver.url("this.txt")
    ret = salt_call_cli.run(
        "--local",
        "file.get_source_sum",
        tmpdir.join("test_source_sum.txt").strpath,
        web_file,
        web_file + ".sha256",
        "verify_ssl={}".format(verify_ssl),
    )
    if not verify_ssl:
        assert (
            ret.json["hsum"]
            == "f2ca1bb6c7e907d06dafe4687e579fce76b37e4e93b7605022da52e6ccc26fd2"
        )
        assert ret.json["hash_type"] == "sha256"
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stderr


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_get_managed_verify_ssl(salt_call_cli, tmpdir, ssl_webserver, verify_ssl):
    """
    test verify_ssl with get_managed
    """
    web_file = ssl_webserver.url("this.txt")
    ret = salt_call_cli.run(
        "--local",
        "file.get_managed",
        tmpdir.join("test_managed.txt").strpath,
        "",
        web_file,
        web_file + ".sha256",
        "",
        "",
        "",
        "",
        "",
        "base",
        "{}",
        "",
        "True",
        "verify_ssl={}".format(verify_ssl),
    )
    if not verify_ssl:
        assert "this.txt" in ret.json[0]
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stdout


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_manage_file_verify_ssl(salt_call_cli, tmpdir, ssl_webserver, verify_ssl):
    """
    test verify_ssl with manage_file
    """
    test_file = tmpdir.join("test_manage_file.txt").strpath
    ret = salt_call_cli.run(
        "--local",
        "file.manage_file",
        test_file,
        "",
        "",
        ssl_webserver.url("this.txt"),
        "{hash_type: 'sha256', 'hsum':"
        " 'f2ca1bb6c7e907d06dafe4687e579fce76b37e4e93b7605022da52e6ccc26fd2'}",
        "",
        "",
        "",
        "",
        "base",
        "",
        "verify_ssl={}".format(verify_ssl),
    )
    if not verify_ssl:
        assert ret.json["changes"] == {"diff": "New file", "mode": "0000"}
        assert ret.json["comment"] == "File {} updated".format(test_file)
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stderr


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_check_managed_changes_verify_ssl(
    salt_call_cli, tmpdir, ssl_webserver, verify_ssl
):
    """
    test verify_ssl with check_managed_changes
    """
    test_file = tmpdir.join("test_managed_changes.txt").strpath
    web_url = ssl_webserver.url("this.txt")
    ret = salt_call_cli.run(
        "--local",
        "file.check_managed_changes",
        test_file,
        web_url,
        web_url + ".sha256",
        "",
        "",
        "",
        "",
        "",
        "jinja",
        "",
        "",
        "base",
        "verify_ssl={}".format(verify_ssl),
    )

    if not verify_ssl:
        assert ret.json["newfile"] == test_file
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stderr


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_check_file_meta_verify_ssl(salt_call_cli, tmpdir, ssl_webserver, verify_ssl):
    """
    test verify_ssl with check_file_meta
    """
    test_file = tmpdir.join("test_check_file_meta.txt").strpath
    with salt.utils.files.fopen(test_file, "w") as fh_:
        fh_.write("test check_file_meta")
    web_url = ssl_webserver.url("this.txt")
    ret = salt_call_cli.run(
        "--local",
        "file.check_file_meta",
        test_file,
        "",
        web_url,
        "{hash_type: 'sha256', 'hsum':"
        " 'f2ca1bb6c7e907d06dafe4687e579fce76b37e4e93b7605022da52e6ccc26fd2'}",
        "",
        "",
        "",
        "",
        "base",
        "verify_ssl={}".format(verify_ssl),
    )

    if not verify_ssl:
        assert (
            len([x for x in ["diff", "user", "group", "mode"] if x in ret.json.keys()])
            == 4
        )
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stderr



"""
Module for Sending Messages via XMPP (a.k.a. Jabber)

.. versionadded:: 2014.1.0

:depends:   - sleekxmpp>=1.3.1
            - pyasn1
            - pyasn1-modules
            - dnspython
:configuration: This module can be used by either passing a jid and password
    directly to send_message, or by specifying the name of a configuration
    profile in the minion config, minion pillar, or master config.

    For example:

    .. code-block:: yaml

        my-xmpp-login:
            xmpp.jid: myuser@jabber.example.org/resourcename
            xmpp.password: verybadpass

    The resourcename refers to the resource that is using this account. It is
    user-definable, and optional. The following configurations are both valid:

    .. code-block:: yaml

        my-xmpp-login:
            xmpp.jid: myuser@jabber.example.org/salt
            xmpp.password: verybadpass

        my-xmpp-login:
            xmpp.jid: myuser@jabber.example.org
            xmpp.password: verybadpass

"""


import logging

HAS_LIBS = False
try:
    from sleekxmpp import ClientXMPP as _ClientXMPP
    from sleekxmpp.exceptions import XMPPError

    HAS_LIBS = True
except ImportError:

    class _ClientXMPP:
        """
        Fake class in order not to raise errors
        """


log = logging.getLogger(__name__)

__virtualname__ = "xmpp"

MUC_DEPRECATED = "Use of send mask waiters is deprecated."


def __virtual__():
    """
    Only load this module if sleekxmpp is installed on this minion.
    """
    if HAS_LIBS:
        return __virtualname__
    return (False, "Module xmpp: required libraries failed to load")


class SleekXMPPMUC(logging.Filter):
    def filter(self, record):
        return not record.getMessage() == MUC_DEPRECATED


class SendMsgBot(_ClientXMPP):
    def __init__(self, jid, password, recipient, msg):
        # PyLint wrongly reports an error when calling super, hence the above
        # disable call
        super().__init__(jid, password)

        self.recipients = [] if recipient is None else [recipient]
        self.rooms = []

        self.msg = msg

        self.add_event_handler("session_start", self.start)

    @classmethod
    def create_multi(
        cls, jid, password, msg, recipients=None, rooms=None, nick="SaltStack Bot"
    ):
        """
        Alternate constructor that accept multiple recipients and rooms
        """
        obj = SendMsgBot(jid, password, None, msg)
        obj.recipients = [] if recipients is None else recipients
        obj.rooms = [] if rooms is None else rooms
        obj.nick = nick
        return obj

    def start(self, event):
        self.send_presence()
        self.get_roster()

        for recipient in self.recipients:
            self.send_message(mto=recipient, mbody=self.msg, mtype="chat")

        for room in self.rooms:
            self.plugin["xep_0045"].joinMUC(room, self.nick, wait=True)
            self.send_message(mto=room, mbody=self.msg, mtype="groupchat")

        self.disconnect(wait=True)


def send_msg(recipient, message, jid=None, password=None, profile=None):
    """
    Send a message to an XMPP recipient. Designed for use in states.

    CLI Examples:

    .. code-block:: bash

        xmpp.send_msg 'admins@xmpp.example.com' 'This is a salt module test' \
            profile='my-xmpp-account'
        xmpp.send_msg 'admins@xmpp.example.com' 'This is a salt module test' \
            jid='myuser@xmpp.example.com/salt' password='verybadpass'
    """
    if profile:
        creds = __salt__["config.option"](profile)
        jid = creds.get("xmpp.jid")
        password = creds.get("xmpp.password")

    xmpp = SendMsgBot(jid, password, recipient, message)
    xmpp.register_plugin("xep_0030")  # Service Discovery
    xmpp.register_plugin("xep_0199")  # XMPP Ping

    if xmpp.connect():
        xmpp.process(block=True)
        return True
    return False


def send_msg_multi(
    message,
    recipients=None,
    rooms=None,
    jid=None,
    password=None,
    nick="SaltStack Bot",
    profile=None,
):
    """
    Send a message to an XMPP recipient, support send message to
    multiple recipients or chat room.

    CLI Examples:

    .. code-block:: bash

        xmpp.send_msg recipients=['admins@xmpp.example.com'] \
            rooms=['secret@conference.xmpp.example.com'] \
            'This is a salt module test' \
            profile='my-xmpp-account'
        xmpp.send_msg recipients=['admins@xmpp.example.com'] \
            rooms=['secret@conference.xmpp.example.com'] \
           'This is a salt module test' \
            jid='myuser@xmpp.example.com/salt' password='verybadpass'

    """

    # Remove: [WARNING ] Use of send mask waiters is deprecated.
    for handler in logging.root.handlers:
        handler.addFilter(SleekXMPPMUC())

    if profile:
        creds = __salt__["config.option"](profile)
        jid = creds.get("xmpp.jid")
        password = creds.get("xmpp.password")

    xmpp = SendMsgBot.create_multi(
        jid, password, message, recipients=recipients, rooms=rooms, nick=nick
    )

    if rooms:
        xmpp.register_plugin("xep_0045")  # MUC plugin
    if xmpp.connect():
        try:
            xmpp.process(block=True)
            return True
        except XMPPError as err:
            log.error("Could not send message, error: %s", err)
    else:
        log.error("Could not connect to XMPP server")
    return False



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import salt.modules.bluez_bluetooth as bluez
import salt.utils.validate.net
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class MockBluetooth:
    """
    Mock class for bluetooth
    """

    def __init__(self):
        pass

    @staticmethod
    def discover_devices(lookup_names):
        """
        Mock method to return all Discoverable devices
        """
        return [["a", "b", "c"], ["d", "e", "f"]]


class BluezTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.bluez
    """

    def setup_loader_modules(self):
        return {bluez: {"bluetooth": MockBluetooth()}}

    def test_version(self):
        """
        Test if return bluetooth version
        """
        mock = MagicMock(return_value="5.7")
        with patch.dict(bluez.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(
                bluez.version(),
                {"PyBluez": "<= 0.18 (Unknown, but installed)", "Bluez": "5.7"},
            )

    def test_address_(self):
        """
        Test of getting address of bluetooth adapter
        """
        mock = MagicMock(return_value="hci : hci0")
        with patch.dict(bluez.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(
                bluez.address_(),
                {"hci ": {"device": "hci ", "path": "/sys/class/bluetooth/hci "}},
            )

    def test_power(self):
        """
        Test of getting address of bluetooth adapter
        """
        mock = MagicMock(return_value={})
        with patch.object(bluez, "address_", mock):
            self.assertRaises(CommandExecutionError, bluez.power, "hci0", "on")

        mock = MagicMock(return_value={"hci0": {"device": "hci0", "power": "on"}})
        with patch.object(bluez, "address_", mock):
            mock = MagicMock(return_value="")
            with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                self.assertTrue(bluez.power("hci0", "on"))

        mock = MagicMock(return_value={"hci0": {"device": "hci0", "power": "on"}})
        with patch.object(bluez, "address_", mock):
            mock = MagicMock(return_value="")
            with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                self.assertFalse(bluez.power("hci0", "off"))

    def test_discoverable(self):
        """
        Test of enabling bluetooth device
        """
        mock = MagicMock(
            side_effect=[
                {},
                {"hci0": {"device": "hci0", "power": "on"}},
                {"hci0": {"device": "hci0", "power": "on"}},
            ]
        )
        with patch.object(bluez, "address_", mock):
            self.assertRaises(CommandExecutionError, bluez.discoverable, "hci0")

            mock = MagicMock(return_value="UP RUNNING ISCAN")
            with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                self.assertTrue(bluez.discoverable("hci0"))

            mock = MagicMock(return_value="")
            with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                self.assertFalse(bluez.discoverable("hci0"))

    def test_noscan(self):
        """
        Test of turning off of scanning modes
        """
        mock = MagicMock(
            side_effect=[
                {},
                {"hci0": {"device": "hci0", "power": "on"}},
                {"hci0": {"device": "hci0", "power": "on"}},
            ]
        )
        with patch.object(bluez, "address_", mock):
            self.assertRaises(CommandExecutionError, bluez.noscan, "hci0")

            mock = MagicMock(return_value="SCAN")
            with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                self.assertFalse(bluez.noscan("hci0"))

            mock = MagicMock(return_value="")
            with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                self.assertTrue(bluez.noscan("hci0"))

    def test_scan(self):
        """
        Test of scanning of bluetooth devices
        """
        self.assertListEqual(bluez.scan(), [{"a": "b"}, {"d": "e"}])

    def test_block(self):
        """
        Test of blocking specific bluetooth device
        """
        mock = MagicMock(side_effect=[False, True])
        with patch.object(salt.utils.validate.net, "mac", mock):
            self.assertRaises(CommandExecutionError, bluez.block, "DE:AD:BE:EF:CA:ZE")

            mock = MagicMock(return_value="")
            with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                self.assertIsNone(bluez.block("DE:AD:BE:EF:CA:FE"))

    def test_unblock(self):
        """
        Test to unblock specific bluetooth device
        """
        mock = MagicMock(side_effect=[False, True])
        with patch.object(salt.utils.validate.net, "mac", mock):
            self.assertRaises(CommandExecutionError, bluez.block, "DE:AD:BE:EF:CA:ZE")

            mock = MagicMock(return_value="")
            with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                self.assertIsNone(bluez.unblock("DE:AD:BE:EF:CA:FE"))

    def test_pair(self):
        """
        Test to pair bluetooth adapter with a device
        """
        mock = MagicMock(side_effect=[False, True, True])
        with patch.object(salt.utils.validate.net, "mac", mock):
            self.assertRaises(
                CommandExecutionError, bluez.pair, "DE:AD:BE:EF:CA:FE", "1234"
            )

            self.assertRaises(
                CommandExecutionError, bluez.pair, "DE:AD:BE:EF:CA:FE", "abcd"
            )

            mock = MagicMock(return_value={"device": "hci0"})
            with patch.object(bluez, "address_", mock):
                mock = MagicMock(return_value="Ok")
                with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                    self.assertListEqual(
                        bluez.pair("DE:AD:BE:EF:CA:FE", "1234"), ["Ok"]
                    )

    def test_unpair(self):
        """
        Test to unpair bluetooth adaptor with a device
        """
        mock = MagicMock(side_effect=[False, True])
        with patch.object(salt.utils.validate.net, "mac", mock):
            self.assertRaises(CommandExecutionError, bluez.unpair, "DE:AD:BE:EF:CA:FE")

            mock = MagicMock(return_value="Ok")
            with patch.dict(bluez.__salt__, {"cmd.run": mock}):
                self.assertListEqual(bluez.unpair("DE:AD:BE:EF:CA:FE"), ["Ok"])

    def test_start(self):
        """
        Test to start bluetooth service
        """
        mock = MagicMock(return_value="Ok")
        with patch.dict(bluez.__salt__, {"service.start": mock}):
            self.assertEqual(bluez.start(), "Ok")

    def test_stop(self):
        """
        Test to stop bluetooth service
        """
        mock = MagicMock(return_value="Ok")
        with patch.dict(bluez.__salt__, {"service.stop": mock}):
            self.assertEqual(bluez.stop(), "Ok")



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import os
import shutil
import uuid

import pytest
import salt.modules.seed as seed
import salt.utils.files
import salt.utils.odict
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SeedTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.seed
    """

    def setup_loader_modules(self):
        return {seed: {}}

    @pytest.mark.slow_test
    def test_mkconfig_odict(self):
        with patch.dict(seed.__opts__, {"master": "foo"}):
            ddd = salt.utils.odict.OrderedDict()
            ddd["b"] = "b"
            ddd["a"] = "b"
            data = seed.mkconfig(ddd, approve_key=False)
            with salt.utils.files.fopen(data["config"]) as fic:
                fdata = fic.read()
                self.assertEqual(fdata, "b: b\na: b\nmaster: foo\n")

    def test_prep_bootstrap(self):
        """
        Test to update and get the random script to a random place
        """
        with patch.dict(
            seed.__salt__,
            {
                "config.gather_bootstrap_script": MagicMock(
                    return_value=os.path.join("BS_PATH", "BS")
                )
            },
        ), patch.object(uuid, "uuid4", return_value="UUID"), patch.object(
            os.path, "exists", return_value=True
        ), patch.object(
            os, "chmod", return_value=None
        ), patch.object(
            shutil, "copy", return_value=None
        ):

            expect = (
                os.path.join("MPT", "tmp", "UUID", "BS"),
                os.sep + os.path.join("tmp", "UUID"),
            )
            self.assertEqual(seed.prep_bootstrap("MPT"), expect)

            expect = (
                os.sep + os.path.join("MPT", "tmp", "UUID", "BS"),
                os.sep + os.path.join("tmp", "UUID"),
            )
            self.assertEqual(seed.prep_bootstrap(os.sep + "MPT"), expect)

    def test_apply_(self):
        """
        Test to seed a location (disk image, directory, or block device)
        with the minion config, approve the minion's key, and/or install
        salt-minion.
        """
        mock = MagicMock(
            side_effect=[
                False,
                {"type": "type", "target": "target"},
                {"type": "type", "target": "target"},
                {"type": "type", "target": "target"},
            ]
        )
        with patch.dict(seed.__salt__, {"file.stats": mock}):
            self.assertEqual(seed.apply_("path"), "path does not exist")

            with patch.object(seed, "_mount", return_value=False):
                self.assertEqual(seed.apply_("path"), "target could not be mounted")

            with patch.object(seed, "_mount", return_value="/mountpoint"):
                with patch.object(os.path, "join", return_value="A"):
                    with patch.object(
                        os, "makedirs", MagicMock(side_effect=OSError("f"))
                    ):
                        with patch.object(os.path, "isdir", return_value=False):
                            self.assertRaises(OSError, seed.apply_, "p")

                    with patch.object(os, "makedirs", MagicMock()):
                        with patch.object(seed, "mkconfig", return_value="A"):
                            with patch.object(
                                seed, "_check_install", return_value=False
                            ):
                                with patch.object(
                                    seed, "_umount", return_value=None
                                ) as umount_mock:
                                    self.assertFalse(seed.apply_("path", install=False))
                                    umount_mock.assert_called_once_with(
                                        "/mountpoint", "target", "type"
                                    )



"""
Test ext_nodes master_tops module
"""


import subprocess
import textwrap

import salt.tops.ext_nodes as ext_nodes
import salt.utils.stringutils
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class ExtNodesTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {
            ext_nodes: {
                "__opts__": {
                    "master_tops": {
                        # Since ext_nodes runs the command with shell=True,
                        # this will keep "command not found" errors from
                        # showing up on the console. We'll be mocking the
                        # communicate results anyway.
                        "ext_nodes": "echo",
                    }
                }
            }
        }

    def test_ext_nodes(self):
        """
        Confirm that subprocess.Popen works as expected and does not raise an
        exception (see https://github.com/saltstack/salt/pull/46863).
        """
        stdout = salt.utils.stringutils.to_bytes(
            textwrap.dedent(
                """\
            classes:
              - one
              - two"""
            )
        )
        communicate_mock = MagicMock(return_value=(stdout, None))
        with patch.object(subprocess.Popen, "communicate", communicate_mock):
            ret = ext_nodes.top(opts={"id": "foo"})
        self.assertEqual(ret, {"base": ["one", "two"]})

    def test_ext_nodes_with_environment(self):
        """
        Same as above, but also tests that the matches are assigned to the proper
        environment if one is returned by the ext_nodes command.
        """
        stdout = salt.utils.stringutils.to_bytes(
            textwrap.dedent(
                """\
            classes:
              - one
              - two
            environment: dev"""
            )
        )
        communicate_mock = MagicMock(return_value=(stdout, None))
        with patch.object(subprocess.Popen, "communicate", communicate_mock):
            ret = ext_nodes.top(opts={"id": "foo"})
        self.assertEqual(ret, {"dev": ["one", "two"]})



"""
Jinja-specific decorators
"""

# Ensure we're using the custom logging from Salt
import salt.log.setup as logging

log = logging.getLogger(__name__)


class JinjaFilter:
    """
    This decorator is used to specify that a function is to be loaded as a
    Jinja filter.
    """

    salt_jinja_filters = {}

    def __init__(self, name=None):
        """ """
        self.name = name

    def __call__(self, function):
        """ """
        name = self.name or function.__name__
        if name not in self.salt_jinja_filters:
            log.debug("Marking '%s' as a jinja filter", name)
            self.salt_jinja_filters[name] = function
        return function


jinja_filter = JinjaFilter


class JinjaTest:
    """
    This decorator is used to specify that a function is to be loaded as a
    Jinja test.
    """

    salt_jinja_tests = {}

    def __init__(self, name=None):
        """ """
        self.name = name

    def __call__(self, function):
        """ """
        name = self.name or function.__name__
        if name not in self.salt_jinja_tests:
            log.debug("Marking '%s' as a jinja test", name)
            self.salt_jinja_tests[name] = function
        return function


jinja_test = JinjaTest


class JinjaGlobal:
    """
    This decorator is used to specify that a function is to be loaded as a
    Jinja global.
    """

    salt_jinja_globals = {}

    def __init__(self, name=None):
        """ """
        self.name = name

    def __call__(self, function):
        """ """
        name = self.name or function.__name__
        if name not in self.salt_jinja_globals:
            log.debug("Marking '%s' as a jinja global", name)
            self.salt_jinja_globals[name] = function
        return function


jinja_global = JinjaGlobal



import pytest


@pytest.fixture(scope="module")
def non_root_account():
    with pytest.helpers.create_account() as account:
        yield account


@pytest.mark.skip_if_not_root
def test_exec_code_all(salt_call_cli, non_root_account):
    ret = salt_call_cli.run(
        "cmd.exec_code_all", "bash", "echo good", runas=non_root_account.username
    )
    assert ret.exitcode == 0



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import salt.states.iptables as iptables
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class IptablesTestCase(TestCase, LoaderModuleMockMixin):
    """
    Validate the iptables state
    """

    def setup_loader_modules(self):
        return {iptables: {}}

    def test_chain_present(self):
        """
        Test to verify the chain is exist.
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

        mock = MagicMock(side_effect=[True, False, False, False])
        with patch.dict(iptables.__salt__, {"iptables.check_chain": mock}):
            ret.update(
                {
                    "comment": (
                        "iptables salt chain is already exist in filter table for ipv4"
                    )
                }
            )
            self.assertDictEqual(iptables.chain_present("salt"), ret)

            with patch.dict(iptables.__opts__, {"test": True}):
                ret.update(
                    {
                        "comment": (
                            "iptables salt chain in filter"
                            " table needs to be set for ipv4"
                        ),
                        "result": None,
                    }
                )
                self.assertDictEqual(iptables.chain_present("salt"), ret)

            with patch.dict(iptables.__opts__, {"test": False}):
                mock = MagicMock(side_effect=[True, ""])
                with patch.dict(iptables.__salt__, {"iptables.new_chain": mock}):
                    ret.update(
                        {
                            "result": True,
                            "comment": (
                                "iptables salt chain in filter"
                                " table create success for ipv4"
                            ),
                            "changes": {"locale": "salt"},
                        }
                    )
                    self.assertDictEqual(iptables.chain_present("salt"), ret)

                    ret.update(
                        {
                            "changes": {},
                            "result": False,
                            "comment": (
                                "Failed to create salt chain in filter table:  for ipv4"
                            ),
                        }
                    )
                    self.assertDictEqual(iptables.chain_present("salt"), ret)

    def test_chain_absent(self):
        """
        Test to verify the chain is absent.
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

        mock = MagicMock(side_effect=[False, True, True, True])
        with patch.dict(iptables.__salt__, {"iptables.check_chain": mock}):
            ret.update(
                {
                    "comment": (
                        "iptables salt chain is already absent in filter table for ipv4"
                    )
                }
            )
            self.assertDictEqual(iptables.chain_absent("salt"), ret)

            with patch.dict(iptables.__opts__, {"test": True}):
                ret.update(
                    {
                        "comment": (
                            "iptables salt chain in filter"
                            " table needs to be removed ipv4"
                        ),
                        "result": None,
                    }
                )
                self.assertDictEqual(iptables.chain_absent("salt"), ret)

            with patch.dict(iptables.__opts__, {"test": False}):
                mock = MagicMock(side_effect=[False, "a"])
                with patch.dict(iptables.__salt__, {"iptables.flush": mock}):
                    mock = MagicMock(return_value=True)
                    with patch.dict(iptables.__salt__, {"iptables.delete_chain": mock}):
                        ret.update(
                            {
                                "changes": {"locale": "salt"},
                                "comment": (
                                    "iptables salt chain in filter"
                                    " table delete success for ipv4"
                                ),
                                "result": True,
                            }
                        )
                        self.assertDictEqual(iptables.chain_absent("salt"), ret)

                    ret.update(
                        {
                            "changes": {},
                            "result": False,
                            "comment": (
                                "Failed to flush salt chain in filter table: a for ipv4"
                            ),
                        }
                    )
                    self.assertDictEqual(iptables.chain_absent("salt"), ret)

    def test_append(self):
        """
        Test to append a rule to a chain
        """
        ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

        self.assertDictEqual(iptables.append("salt", rules=[]), ret)

        mock = MagicMock(return_value=[])
        with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
            mock = MagicMock(return_value="a")
            with patch.dict(iptables.__salt__, {"iptables.build_rule": mock}):
                mock = MagicMock(side_effect=[True, False, False, False, False, True])
                with patch.dict(iptables.__salt__, {"iptables.check": mock}):
                    ret.update(
                        {
                            "comment": (
                                "iptables rule for salt already set (a) for ipv4"
                            ),
                            "result": True,
                        }
                    )
                    self.assertDictEqual(
                        iptables.append("salt", table="", chain=""), ret
                    )

                    with patch.dict(iptables.__opts__, {"test": True}):
                        ret.update(
                            {
                                "result": None,
                                "comment": (
                                    "iptables rule for salt"
                                    " needs to be set (a) for ipv4"
                                ),
                            }
                        )
                        self.assertDictEqual(
                            iptables.append("salt", table="", chain=""), ret
                        )

                    with patch.dict(iptables.__opts__, {"test": False}):
                        mock = MagicMock(side_effect=[True, False, True, True])
                        with patch.dict(iptables.__salt__, {"iptables.append": mock}):
                            ret.update(
                                {
                                    "changes": {"locale": "salt"},
                                    "result": True,
                                    "comment": (
                                        "Set iptables rule for salt to: a for ipv4"
                                    ),
                                }
                            )
                            self.assertDictEqual(
                                iptables.append("salt", table="", chain=""), ret
                            )

                            ret.update(
                                {
                                    "changes": {},
                                    "result": False,
                                    "comment": (
                                        "Failed to set iptables"
                                        " rule for salt.\nAttempted rule was"
                                        " a for ipv4"
                                    ),
                                }
                            )
                            self.assertDictEqual(
                                iptables.append("salt", table="", chain=""), ret
                            )

                            mock_save = MagicMock(
                                side_effect=['Wrote 1 lines to "/tmp/iptables"', ""]
                            )
                            with patch.dict(
                                iptables.__salt__, {"iptables.save": mock_save}
                            ):
                                mock_get_saved_rules = MagicMock(side_effect=[""])
                                with patch.dict(
                                    iptables.__salt__,
                                    {"iptables.get_saved_rules": mock_get_saved_rules},
                                ):
                                    mock = MagicMock(side_effect=[""])
                                    with patch.dict(
                                        iptables.__salt__, {"iptables.get_rules": mock}
                                    ):
                                        ret.update(
                                            {
                                                "changes": {"locale": "salt"},
                                                "result": True,
                                                "comment": "Set and saved iptables rule"
                                                ' salt for ipv4\na\nWrote 1 lines to "/tmp/iptables"',
                                            }
                                        )
                                        self.assertDictEqual(
                                            iptables.append(
                                                "salt",
                                                table="",
                                                chain="",
                                                save="/tmp/iptables",
                                            ),
                                            ret,
                                        )
                                        ret.update(
                                            {
                                                "changes": {},
                                                "result": True,
                                                "comment": "iptables rule for salt already set (a) for ipv4",
                                            }
                                        )
                                        self.assertDictEqual(
                                            iptables.append(
                                                "salt",
                                                table="",
                                                chain="",
                                                save="/tmp/iptables",
                                            ),
                                            ret,
                                        )
                                        self.assertEqual(
                                            mock_get_saved_rules.mock_calls[0][2][
                                                "conf_file"
                                            ],
                                            "/tmp/iptables",
                                        )
                                        self.assertEqual(
                                            mock_save.mock_calls[0][2]["filename"],
                                            "/tmp/iptables",
                                        )

    def test_insert(self):
        """
        Test to insert a rule into a chain
        """
        ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

        self.assertDictEqual(iptables.insert("salt", rules=[]), ret)

        mock = MagicMock(return_value=[])
        with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
            mock = MagicMock(return_value="a")
            with patch.dict(iptables.__salt__, {"iptables.build_rule": mock}):
                mock = MagicMock(side_effect=[True, False, False, False, False, True])
                with patch.dict(iptables.__salt__, {"iptables.check": mock}):
                    ret.update(
                        {
                            "comment": (
                                "iptables rule for salt already set for ipv4 (a)"
                            ),
                            "result": True,
                        }
                    )
                    self.assertDictEqual(
                        iptables.insert("salt", table="", chain=""), ret
                    )

                    with patch.dict(iptables.__opts__, {"test": True}):
                        ret.update(
                            {
                                "result": None,
                                "comment": (
                                    "iptables rule for salt"
                                    " needs to be set for ipv4 (a)"
                                ),
                            }
                        )
                        self.assertDictEqual(
                            iptables.insert("salt", table="", chain=""), ret
                        )

                    with patch.dict(iptables.__opts__, {"test": False}):
                        mock = MagicMock(side_effect=[False, True, False, True])
                        with patch.dict(iptables.__salt__, {"iptables.insert": mock}):
                            ret.update(
                                {
                                    "changes": {"locale": "salt"},
                                    "result": True,
                                    "comment": (
                                        "Set iptables rule for salt to: a for ipv4"
                                    ),
                                }
                            )
                            self.assertDictEqual(
                                iptables.insert(
                                    "salt", table="", chain="", position=""
                                ),
                                ret,
                            )

                            ret.update(
                                {
                                    "changes": {},
                                    "result": False,
                                    "comment": (
                                        "Failed to set iptables"
                                        " rule for salt.\nAttempted rule was a"
                                    ),
                                }
                            )
                            self.assertDictEqual(
                                iptables.insert(
                                    "salt", table="", chain="", position=""
                                ),
                                ret,
                            )

                            mock_save = MagicMock(
                                side_effect=['Wrote 1 lines to "/tmp/iptables"', ""]
                            )
                            with patch.dict(
                                iptables.__salt__, {"iptables.save": mock_save}
                            ):
                                mock_get_saved_rules = MagicMock(side_effect=[""])
                                with patch.dict(
                                    iptables.__salt__,
                                    {"iptables.get_saved_rules": mock_get_saved_rules},
                                ):
                                    mock = MagicMock(side_effect=[""])
                                    with patch.dict(
                                        iptables.__salt__, {"iptables.get_rules": mock}
                                    ):
                                        ret.update(
                                            {
                                                "changes": {"locale": "salt"},
                                                "result": True,
                                                "comment": "Set and saved iptables rule"
                                                ' salt for ipv4\na\nWrote 1 lines to "/tmp/iptables"',
                                            }
                                        )
                                        self.assertDictEqual(
                                            iptables.insert(
                                                "salt",
                                                table="",
                                                chain="",
                                                position="",
                                                save="/tmp/iptables",
                                            ),
                                            ret,
                                        )
                                        ret.update(
                                            {
                                                "changes": {},
                                                "result": True,
                                                "comment": "iptables rule for salt already set for ipv4 (a)",
                                            }
                                        )
                                        self.assertDictEqual(
                                            iptables.insert(
                                                "salt",
                                                table="",
                                                chain="",
                                                position="",
                                                save="/tmp/iptables",
                                            ),
                                            ret,
                                        )
                                        self.assertEqual(
                                            mock_get_saved_rules.mock_calls[0][2][
                                                "conf_file"
                                            ],
                                            "/tmp/iptables",
                                        )
                                        self.assertEqual(
                                            mock_save.mock_calls[0][2]["filename"],
                                            "/tmp/iptables",
                                        )

    def test_delete(self):
        """
        Test to delete a rule to a chain
        """
        ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

        self.assertDictEqual(iptables.delete("salt", rules=[]), ret)

        mock = MagicMock(return_value=[])
        with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
            mock = MagicMock(return_value="a")
            with patch.dict(iptables.__salt__, {"iptables.build_rule": mock}):
                mock = MagicMock(side_effect=[False, True, True, True, True, False])
                with patch.dict(iptables.__salt__, {"iptables.check": mock}):
                    ret.update(
                        {
                            "comment": (
                                "iptables rule for salt already absent for ipv4 (a)"
                            ),
                            "result": True,
                        }
                    )
                    self.assertDictEqual(
                        iptables.delete("salt", table="", chain=""), ret
                    )

                    with patch.dict(iptables.__opts__, {"test": True}):
                        ret.update(
                            {
                                "result": None,
                                "comment": (
                                    "iptables rule for salt needs"
                                    " to be deleted for ipv4 (a)"
                                ),
                            }
                        )
                        self.assertDictEqual(
                            iptables.delete("salt", table="", chain=""), ret
                        )

                    with patch.dict(iptables.__opts__, {"test": False}):
                        mock = MagicMock(side_effect=[False, True, False, False])
                        with patch.dict(iptables.__salt__, {"iptables.delete": mock}):
                            ret.update(
                                {
                                    "result": True,
                                    "changes": {"locale": "salt"},
                                    "comment": "Delete iptables rule for salt a",
                                }
                            )
                            self.assertDictEqual(
                                iptables.delete(
                                    "salt", table="", chain="", position=""
                                ),
                                ret,
                            )

                            ret.update(
                                {
                                    "result": False,
                                    "changes": {},
                                    "comment": (
                                        "Failed to delete iptables"
                                        " rule for salt.\nAttempted rule was a"
                                    ),
                                }
                            )
                            self.assertDictEqual(
                                iptables.delete(
                                    "salt", table="", chain="", position=""
                                ),
                                ret,
                            )

                            mock_save = MagicMock(
                                side_effect=['Wrote 1 lines to "/tmp/iptables"', ""]
                            )
                            with patch.dict(
                                iptables.__salt__, {"iptables.save": mock_save}
                            ):
                                mock = MagicMock(side_effect=[True, False])
                                with patch.dict(
                                    iptables.__salt__, {"iptables.check": mock}
                                ):
                                    mock = MagicMock(side_effect=[""])
                                    with patch.dict(
                                        iptables.__salt__, {"iptables.get_rules": mock}
                                    ):
                                        ret.update(
                                            {
                                                "changes": {"locale": "salt"},
                                                "result": True,
                                                "comment": "Deleted and saved iptables rule"
                                                ' salt for ipv4\na\nWrote 1 lines to "/tmp/iptables"',
                                            }
                                        )
                                        self.assertDictEqual(
                                            iptables.delete(
                                                "salt",
                                                table="",
                                                chain="",
                                                save="/tmp/iptables",
                                            ),
                                            ret,
                                        )
                                        ret.update(
                                            {
                                                "changes": {},
                                                "result": True,
                                                "comment": "iptables rule for salt already absent for ipv4 (a)",
                                            }
                                        )
                                        self.assertDictEqual(
                                            iptables.delete(
                                                "salt",
                                                table="",
                                                chain="",
                                                save="/tmp/iptables",
                                            ),
                                            ret,
                                        )
                                        self.assertEqual(
                                            mock_save.mock_calls[0][2]["filename"],
                                            "/tmp/iptables",
                                        )

    def test_set_policy(self):
        """
        Test to sets the default policy for iptables firewall tables
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

        mock = MagicMock(return_value=[])
        with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
            mock = MagicMock(return_value="stack")
            with patch.dict(iptables.__salt__, {"iptables.get_policy": mock}):
                ret.update(
                    {
                        "comment": (
                            "iptables default policy for chain"
                            "  on table  for ipv4 already set to stack"
                        )
                    }
                )
                self.assertDictEqual(
                    iptables.set_policy("salt", table="", chain="", policy="stack"), ret
                )

                with patch.dict(iptables.__opts__, {"test": True}):
                    ret.update(
                        {
                            "comment": (
                                "iptables default policy for chain"
                                "  on table  for ipv4 needs to be set"
                                " to sal"
                            ),
                            "result": None,
                        }
                    )
                    self.assertDictEqual(
                        iptables.set_policy("salt", table="", chain="", policy="sal"),
                        ret,
                    )

                with patch.dict(iptables.__opts__, {"test": False}):
                    mock = MagicMock(side_effect=[False, True])
                    with patch.dict(iptables.__salt__, {"iptables.set_policy": mock}):
                        ret.update(
                            {
                                "changes": {"locale": "salt"},
                                "comment": "Set default policy for  to sal family ipv4",
                                "result": True,
                            }
                        )
                        self.assertDictEqual(
                            iptables.set_policy(
                                "salt", table="", chain="", policy="sal"
                            ),
                            ret,
                        )

                        ret.update(
                            {
                                "comment": "Failed to set iptables default policy",
                                "result": False,
                                "changes": {},
                            }
                        )
                        self.assertDictEqual(
                            iptables.set_policy(
                                "salt", table="", chain="", policy="sal"
                            ),
                            ret,
                        )

    def test_flush(self):
        """
        Test to flush current iptables state
        """
        ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

        mock = MagicMock(return_value=[])
        with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
            with patch.dict(iptables.__opts__, {"test": True}):
                ret.update(
                    {
                        "comment": (
                            "iptables rules in salt table filter"
                            " chain ipv4 family needs to be flushed"
                        )
                    }
                )
                self.assertDictEqual(iptables.flush("salt"), ret)

            with patch.dict(iptables.__opts__, {"test": False}):
                mock = MagicMock(side_effect=[False, True])
                with patch.dict(iptables.__salt__, {"iptables.flush": mock}):
                    ret.update(
                        {
                            "changes": {"locale": "salt"},
                            "comment": (
                                "Flush iptables rules in  table  chain ipv4 family"
                            ),
                            "result": True,
                        }
                    )
                    self.assertDictEqual(
                        iptables.flush("salt", table="", chain=""), ret
                    )

                    ret.update(
                        {
                            "changes": {},
                            "comment": "Failed to flush iptables rules",
                            "result": False,
                        }
                    )
                    self.assertDictEqual(
                        iptables.flush("salt", table="", chain=""), ret
                    )

    def test_mod_aggregate(self):
        """
        Test to mod_aggregate function
        """
        self.assertDictEqual(
            iptables.mod_aggregate({"fun": "salt"}, [], []), {"fun": "salt"}
        )

        self.assertDictEqual(
            iptables.mod_aggregate({"fun": "append"}, [], []), {"fun": "append"}
        )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.states.artifactory as artifactory
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {artifactory: {}}


def test_downloaded():
    """
    Test to ensures that the artifact from artifactory exists at
    given location.
    """
    name = "jboss"
    arti_url = "http://artifactory.intranet.example.com/artifactory"
    artifact = {
        "artifactory_url": arti_url,
        "artifact_id": "module",
        "repository": "libs-release-local",
        "packaging": "jar",
        "group_id": "com.company.module",
        "classifier": "sources",
        "version": "1.0",
    }

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mck = MagicMock(return_value={"status": False, "changes": {}, "comment": ""})
    with patch.dict(artifactory.__salt__, {"artifactory.get_release": mck}):
        assert artifactory.downloaded(name, artifact) == ret

    with patch.object(
        artifactory,
        "__fetch_from_artifactory",
        MagicMock(side_effect=Exception("error")),
    ):
        ret = artifactory.downloaded(name, artifact)
        assert ret["result"] is False
        assert ret["comment"] == "error"


def test_downloaded_test_true():
    """
    Test to ensures that the artifact from artifactory exists at
    given location.
    """
    name = "jboss"
    arti_url = "http://artifactory.intranet.example.com/artifactory"
    artifact = {
        "artifactory_url": arti_url,
        "artifact_id": "module",
        "repository": "libs-release-local",
        "packaging": "jar",
        "group_id": "com.company.module",
        "classifier": "sources",
        "version": "1.0",
    }

    ret = {
        "name": name,
        "result": True,
        "changes": {},
        "comment": (
            "Artifact would be downloaded from URL:"
            " http://artifactory.intranet.example.com/artifactory"
        ),
    }

    mck = MagicMock(return_value={"status": False, "changes": {}, "comment": ""})
    with patch.dict(artifactory.__salt__, {"artifactory.get_release": mck}):
        with patch.dict(artifactory.__opts__, {"test": True}):
            assert artifactory.downloaded(name, artifact) == ret

    with patch.object(
        artifactory,
        "__fetch_from_artifactory",
        MagicMock(side_effect=Exception("error")),
    ):
        ret = artifactory.downloaded(name, artifact)
        assert ret["result"] is False
        assert ret["comment"] == "error"



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)

    salt.config.schemas.minion
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Minion configuration schema
"""


from salt.config.schemas.common import IncludeConfig, MinionDefaultInclude
from salt.utils.schema import IPv4Item, Schema

# XXX: THIS IS WAY TOO MINIMAL, BUT EXISTS TO IMPLEMENT salt-ssh


class MinionConfiguration(Schema):

    # Because salt's configuration is very permissive with additioal
    # configuration settings, let's allow them in the schema or validation
    # would fail
    __allow_additional_items__ = True

    interface = IPv4Item(title="Interface")

    default_include = MinionDefaultInclude()
    include = IncludeConfig()



import sys
import textwrap

import pytest
import salt.modules.state as state
import salt.state
import salt.utils.files
import salt.utils.json
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skipif(
        sys.version_info < (3, 6), reason="Dictionaries are not ordered under Py3.5"
    ),
]


@pytest.fixture
def configure_loader_modules(salt_minion_factory):
    return {
        state: {
            "__opts__": salt_minion_factory.config.copy(),
            "__salt__": {"saltutil.is_running": MagicMock(return_value=[])},
        },
    }


@pytest.fixture
def cachedir(tmp_path):
    path = tmp_path / "cache"
    path.mkdir()
    return path


@pytest.fixture
def fileserver_root(tmp_path):
    path = tmp_path / "fileserver-root"
    path.mkdir()
    return path


@pytest.fixture
def saltenvs():
    return ["base", "foo", "bar", "baz"]


@pytest.fixture
def saltenv_roots(fileserver_root, saltenvs):
    return {env: fileserver_root / env for env in saltenvs}


@pytest.fixture
def base_top_file(saltenv_roots):
    return str(saltenv_roots["base"] / "top.sls")


@pytest.fixture
def dunder_opts(saltenv_roots, saltenvs):
    return {
        "file_client": "local",
        "default_top": "base",
        "env_order": saltenv_roots,
        "file_roots": {
            "base": [str(saltenv_roots["base"])],
            "foo": [str(saltenv_roots["foo"])],
            "bar": [str(saltenv_roots["bar"])],
            "baz": [str(saltenv_roots["baz"])],
        },
    }


@pytest.fixture(autouse=True)
def state_tree(saltenv_roots, saltenvs):
    # Write top files for all but the "baz" environment
    for env, path in saltenv_roots.items():
        path.mkdir()
        if env == "baz":
            continue
        top_file = path / "top.sls"
        with salt.utils.files.fopen(str(top_file), "w") as fp_:
            # Add a section for every environment to each top file, with
            # the SLS target prefixed with the current saltenv.
            for env_name in saltenvs:
                fp_.write(
                    textwrap.dedent(
                        """\
                    {env_name}:
                      '*':
                        - {saltenv}_{env_name}
                    """.format(
                            env_name=env_name, saltenv=env
                        )
                    )
                )


@pytest.fixture
def limited_base_top_file(state_tree, base_top_file):
    with salt.utils.files.fopen(base_top_file, "w") as fp_:
        fp_.write(
            textwrap.dedent(
                """\
            base:
              '*':
                - base_base
            """
            )
        )


def show_top(dunder_opts, **kwargs):
    dunder_opts.update(kwargs)
    with patch.dict(state.__opts__, dunder_opts), patch.object(
        salt.state.State, "_gather_pillar", MagicMock(return_value={})
    ):
        ret = state.show_top()
        return ret


def test_merge_strategy_merge(dunder_opts):
    """
    Base overrides everything
    """
    ret = show_top(dunder_opts, top_file_merging_strategy="merge")
    assert ret == {
        "base": ["base_base"],
        "foo": ["base_foo"],
        "bar": ["base_bar"],
        "baz": ["base_baz"],
    }


@pytest.mark.usefixtures("limited_base_top_file")
def test_merge_strategy_merge_limited_base(dunder_opts, base_top_file):
    """
    Test with a "base" top file containing only a "base" section. The "baz"
    saltenv should not be in the return data because that env doesn't have
    its own top file and there will be no "baz" section in the "base" env's
    top file.

    Next, append a "baz" section to the rewritten top file and we should
    get results for that saltenv in the return data.
    """
    ret = show_top(dunder_opts, top_file_merging_strategy="merge")
    assert ret == {
        "base": ["base_base"],
        "foo": ["foo_foo"],
        "bar": ["bar_bar"],
    }

    # Add a "baz" section
    with salt.utils.files.fopen(base_top_file, "a") as fp_:
        fp_.write(
            textwrap.dedent(
                """\
            baz:
              '*':
                - base_baz
            """
            )
        )

    ret = show_top(dunder_opts, top_file_merging_strategy="merge")
    assert ret == {
        "base": ["base_base"],
        "foo": ["foo_foo"],
        "bar": ["bar_bar"],
        "baz": ["base_baz"],
    }


def test_merge_strategy_merge_state_top_saltenv_base(dunder_opts):
    """
    This tests with state_top_saltenv=base, which should pull states *only*
    from the base saltenv.
    """
    ret = show_top(
        dunder_opts, top_file_merging_strategy="merge", state_top_saltenv="base"
    )
    assert ret == {
        "base": ["base_base"],
        "foo": ["base_foo"],
        "bar": ["base_bar"],
        "baz": ["base_baz"],
    }


def test_merge_strategy_merge_state_top_saltenv_foo(dunder_opts):
    """
    This tests with state_top_saltenv=foo, which should pull states *only*
    from the foo saltenv. Since that top file is only authoritative for
    its own saltenv, *only* the foo saltenv's matches from the foo top file
    should be in the return data.
    """
    ret = show_top(
        dunder_opts, top_file_merging_strategy="merge", state_top_saltenv="foo"
    )
    assert ret == {"foo": ["foo_foo"]}


def test_merge_strategy_merge_all(dunder_opts):
    """
    Include everything in every top file
    """
    ret = show_top(dunder_opts, top_file_merging_strategy="merge_all")
    assert ret == {
        "base": ["base_base", "foo_base", "bar_base"],
        "foo": ["base_foo", "foo_foo", "bar_foo"],
        "bar": ["base_bar", "foo_bar", "bar_bar"],
        "baz": ["base_baz", "foo_baz", "bar_baz"],
    }


def test_merge_strategy_merge_all_alternate_env_order(dunder_opts):
    """
    Use an alternate env_order. This should change the order in which the
    SLS targets appear in the result.
    """
    ret = show_top(
        dunder_opts,
        top_file_merging_strategy="merge_all",
        env_order=["bar", "foo", "base"],
    )
    assert ret == {
        "base": ["bar_base", "foo_base", "base_base"],
        "foo": ["bar_foo", "foo_foo", "base_foo"],
        "bar": ["bar_bar", "foo_bar", "base_bar"],
        "baz": ["bar_baz", "foo_baz", "base_baz"],
    }


def test_merge_strategy_merge_all_state_top_saltenv_base(dunder_opts):
    """
    This tests with state_top_saltenv=base, which should pull states *only*
    from the base saltenv. Since we are using the "merge_all" strategy, all
    the states from that top file should be in the return data.
    """
    ret = show_top(
        dunder_opts, top_file_merging_strategy="merge_all", state_top_saltenv="base"
    )
    assert ret == {
        "base": ["base_base"],
        "foo": ["base_foo"],
        "bar": ["base_bar"],
        "baz": ["base_baz"],
    }


def test_merge_strategy_merge_all_state_top_saltenv_foo(dunder_opts):
    """
    This tests with state_top_saltenv=foo, which should pull states *only*
    from the foo saltenv. Since we are using the "merge_all" strategy, all
    the states from that top file should be in the return data.
    """
    ret = show_top(
        dunder_opts, top_file_merging_strategy="merge_all", state_top_saltenv="foo"
    )
    assert ret == {
        "base": ["foo_base"],
        "foo": ["foo_foo"],
        "bar": ["foo_bar"],
        "baz": ["foo_baz"],
    }


def test_merge_strategy_same(dunder_opts):
    """
    Each env should get its SLS targets from its own top file, with the
    "baz" env pulling from "base" since default_top=base and there is no
    top file in the "baz" saltenv.
    """
    ret = show_top(dunder_opts, top_file_merging_strategy="same")
    assert ret == {
        "base": ["base_base"],
        "foo": ["foo_foo"],
        "bar": ["bar_bar"],
        "baz": ["base_baz"],
    }


@pytest.mark.usefixtures("limited_base_top_file")
def test_merge_strategy_same_limited_base(dunder_opts):
    """
    Each env should get its SLS targets from its own top file, with the
    "baz" env pulling from "base" since default_top=base and there is no
    top file in the "baz" saltenv.
    """
    ret = show_top(dunder_opts, top_file_merging_strategy="same")
    assert ret == {
        "base": ["base_base"],
        "foo": ["foo_foo"],
        "bar": ["bar_bar"],
    }


def test_merge_strategy_same_default_top_foo(dunder_opts):
    """
    Each env should get its SLS targets from its own top file, with the
    "baz" env pulling from "foo" since default_top=foo and there is no top
    file in the "baz" saltenv.
    """
    ret = show_top(dunder_opts, top_file_merging_strategy="same", default_top="foo")
    assert ret == {
        "base": ["base_base"],
        "foo": ["foo_foo"],
        "bar": ["bar_bar"],
        "baz": ["foo_baz"],
    }


def test_merge_strategy_same_state_top_saltenv_base(dunder_opts):
    """
    Test the state_top_saltenv parameter to load states exclusively from
    the base saltenv, with the "same" merging strategy. This should
    result in just the base environment's states from the base top file
    being in the merged result.
    """
    ret = show_top(
        dunder_opts, top_file_merging_strategy="same", state_top_saltenv="base"
    )
    assert ret == {"base": ["base_base"]}


def test_merge_strategy_same_state_top_saltenv_foo(dunder_opts):
    """
    Test the state_top_saltenv parameter to load states exclusively from
    the foo saltenv, with the "same" merging strategy. This should
    result in just the foo environment's states from the foo top file
    being in the merged result.
    """
    ret = show_top(
        dunder_opts, top_file_merging_strategy="same", state_top_saltenv="foo"
    )
    assert ret == {"foo": ["foo_foo"]}


def test_merge_strategy_same_state_top_saltenv_baz(dunder_opts):
    """
    Test the state_top_saltenv parameter to load states exclusively from
    the baz saltenv, with the "same" merging strategy. This should
    result in an empty dictionary since there is no top file in that
    environment.
    """
    ret = show_top(
        dunder_opts, top_file_merging_strategy="same", state_top_saltenv="baz"
    )
    assert ret == {}



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import salt.modules.http as http
import salt.utils.http
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class HttpTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.http
    """

    def setup_loader_modules(self):
        return {http: {}}

    def test_query(self):
        """
        Test for Query a resource, and decode the return data
        """
        with patch.object(salt.utils.http, "query", return_value="A"):
            self.assertEqual(http.query("url"), "A")

    def test_wait_for_with_interval(self):
        """
        Test for wait_for_successful_query waits for request_interval
        """

        query_mock = MagicMock(side_effect=[{"error": "error"}, {}])

        with patch.object(salt.utils.http, "query", query_mock):
            with patch("time.sleep", MagicMock()) as sleep_mock:
                self.assertEqual(
                    http.wait_for_successful_query("url", request_interval=1), {}
                )
                sleep_mock.assert_called_once_with(1)

    def test_wait_for_without_interval(self):
        """
        Test for wait_for_successful_query waits for request_interval
        """

        query_mock = MagicMock(side_effect=[{"error": "error"}, {}])

        with patch.object(salt.utils.http, "query", query_mock):
            with patch("time.sleep", MagicMock()) as sleep_mock:
                self.assertEqual(http.wait_for_successful_query("url"), {})
                sleep_mock.assert_not_called()



"""
    :codeauthor: :email:`Alexandru Bleotu <alexandru.bleotu@morganstanley.com>`

    Tests for functions in salt.modules.esxdatacenter
"""

import salt.modules.esxdatacenter as esxdatacenter
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class GetDetailsTestCase(TestCase, LoaderModuleMockMixin):
    """Tests for salt.modules.esxdatacenter.get_details"""

    def setup_loader_modules(self):
        return {esxdatacenter: {"__proxy__": {}}}

    def test_get_details(self):
        mock_get_details = MagicMock()
        with patch.dict(
            esxdatacenter.__proxy__, {"esxdatacenter.get_details": mock_get_details}
        ):
            esxdatacenter.get_details()
        mock_get_details.assert_called_once_with()



"""
Amazon S3 Fileserver Backend

.. versionadded:: 0.16.0

This backend exposes directories in S3 buckets as Salt environments. To enable
this backend, add ``s3fs`` to the :conf_master:`fileserver_backend` option in the
Master config file.

.. code-block:: yaml

    fileserver_backend:
      - s3fs

S3 credentials must also be set in the master config file:

.. code-block:: yaml

    s3.keyid: GKTADJGHEIQSXMKKRBJ08H
    s3.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

Alternatively, if on EC2 these credentials can be automatically loaded from
instance metadata.

This fileserver supports two modes of operation for the buckets:

1. :strong:`A single bucket per environment`

   .. code-block:: yaml

    s3.buckets:
      production:
        - bucket1
        - bucket2
      staging:
        - bucket3
        - bucket4

2. :strong:`Multiple environments per bucket`

   .. code-block:: yaml

    s3.buckets:
      - bucket1
      - bucket2
      - bucket3
      - bucket4

Note that bucket names must be all lowercase both in the AWS console and in
Salt, otherwise you may encounter ``SignatureDoesNotMatch`` errors.

A multiple-environment bucket must adhere to the following root directory
structure::

    s3://<bucket name>/<environment>/<files>

.. note:: This fileserver back-end requires the use of the MD5 hashing algorithm.
    MD5 may not be compliant with all security policies.

.. note:: This fileserver back-end is only compatible with MD5 ETag hashes in
    the S3 metadata. This means that you must use SSE-S3 or plaintext for
    bucket encryption, and that you must not use multipart upload when
    uploading to your bucket. More information here:
    https://docs.aws.amazon.com/AmazonS3/latest/API/RESTCommonResponseHeaders.html

    Objects without an MD5 ETag will be fetched on every fileserver update.

    If you deal with objects greater than 8MB, then you should use the
    following AWS CLI config to avoid mutipart upload:

    .. code-block:: text

        s3 =
          multipart_threshold = 1024MB

    More info here:
    https://docs.aws.amazon.com/cli/latest/topic/s3-config.html
"""


import datetime
import logging
import os
import pickle
import time
import urllib.parse

import salt.fileserver as fs
import salt.modules
import salt.utils.files
import salt.utils.gzip_util
import salt.utils.hashutils
import salt.utils.versions

log = logging.getLogger(__name__)

S3_CACHE_EXPIRE = 30  # cache for 30 seconds
S3_SYNC_ON_UPDATE = True  # sync cache on update rather than jit


def envs():
    """
    Return a list of directories within the bucket that can be
    used as environments.
    """

    # update and grab the envs from the metadata keys
    metadata = _init()
    return list(metadata.keys())


def update():
    """
    Update the cache file for the bucket.
    """

    metadata = _init()

    if S3_SYNC_ON_UPDATE:
        # sync the buckets to the local cache
        log.info("Syncing local cache from S3...")
        for saltenv, env_meta in metadata.items():
            for bucket_files in _find_files(env_meta):
                for bucket, files in bucket_files.items():
                    for file_path in files:
                        cached_file_path = _get_cached_file_name(
                            bucket, saltenv, file_path
                        )
                        log.info("%s - %s : %s", bucket, saltenv, file_path)

                        # load the file from S3 if it's not in the cache or it's old
                        _get_file_from_s3(
                            metadata, saltenv, bucket, file_path, cached_file_path
                        )

        log.info("Sync local cache from S3 completed.")


def find_file(path, saltenv="base", **kwargs):
    """
    Look through the buckets cache file for a match.
    If the field is found, it is retrieved from S3 only if its cached version
    is missing, or if the MD5 does not match.
    """
    if "env" in kwargs:
        # "env" is not supported; Use "saltenv".
        kwargs.pop("env")

    fnd = {"bucket": None, "path": None}

    metadata = _init()
    if not metadata or saltenv not in metadata:
        return fnd

    env_files = _find_files(metadata[saltenv])

    if not _is_env_per_bucket():
        path = os.path.join(saltenv, path)

    # look for the files and check if they're ignored globally
    for bucket in env_files:
        for bucket_name, files in bucket.items():
            if path in files and not fs.is_file_ignored(__opts__, path):
                fnd["bucket"] = bucket_name
                fnd["path"] = path
                break
        else:
            continue  # only executes if we didn't break
        break

    if not fnd["path"] or not fnd["bucket"]:
        return fnd

    cached_file_path = _get_cached_file_name(fnd["bucket"], saltenv, path)

    # jit load the file from S3 if it's not in the cache or it's old
    _get_file_from_s3(metadata, saltenv, fnd["bucket"], path, cached_file_path)

    return fnd


def file_hash(load, fnd):
    """
    Return an MD5 file hash
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    ret = {}

    if "saltenv" not in load:
        return ret

    if "path" not in fnd or "bucket" not in fnd or not fnd["path"]:
        return ret

    cached_file_path = _get_cached_file_name(
        fnd["bucket"], load["saltenv"], fnd["path"]
    )

    if os.path.isfile(cached_file_path):
        ret["hsum"] = salt.utils.hashutils.get_hash(cached_file_path)
        ret["hash_type"] = "md5"

    return ret


def serve_file(load, fnd):
    """
    Return a chunk from a file based on the data received
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    ret = {"data": "", "dest": ""}

    if "path" not in load or "loc" not in load or "saltenv" not in load:
        return ret

    if "path" not in fnd or "bucket" not in fnd:
        return ret

    gzip = load.get("gzip", None)

    # get the saltenv/path file from the cache
    cached_file_path = _get_cached_file_name(
        fnd["bucket"], load["saltenv"], fnd["path"]
    )

    ret["dest"] = _trim_env_off_path([fnd["path"]], load["saltenv"])[0]

    with salt.utils.files.fopen(cached_file_path, "rb") as fp_:
        fp_.seek(load["loc"])
        data = fp_.read(__opts__["file_buffer_size"])
        if data and not salt.utils.files.is_binary(cached_file_path):
            data = data.decode(__salt_system_encoding__)
        if gzip and data:
            data = salt.utils.gzip_util.compress(data, gzip)
            ret["gzip"] = gzip
        ret["data"] = data
    return ret


def file_list(load):
    """
    Return a list of all files on the file server in a specified environment
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    ret = []

    if "saltenv" not in load:
        return ret

    saltenv = load["saltenv"]
    metadata = _init()

    if not metadata or saltenv not in metadata:
        return ret
    for bucket in _find_files(metadata[saltenv]):
        for buckets in bucket.values():
            files = [f for f in buckets if not fs.is_file_ignored(__opts__, f)]
            ret += _trim_env_off_path(files, saltenv)

    return ret


def file_list_emptydirs(load):
    """
    Return a list of all empty directories on the master
    """
    # TODO - implement this
    _init()

    return []


def dir_list(load):
    """
    Return a list of all directories on the master
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    ret = []

    if "saltenv" not in load:
        return ret

    saltenv = load["saltenv"]
    metadata = _init()

    if not metadata or saltenv not in metadata:
        return ret

    # grab all the dirs from the buckets cache file
    for bucket in _find_dirs(metadata[saltenv]):
        for dirs in bucket.values():
            # trim env and trailing slash
            dirs = _trim_env_off_path(dirs, saltenv, trim_slash=True)
            # remove empty string left by the base env dir in single bucket mode
            ret += [_f for _f in dirs if _f]

    return ret


def _get_s3_key():
    """
    Get AWS keys from pillar or config
    """

    key = __opts__["s3.key"] if "s3.key" in __opts__ else None
    keyid = __opts__["s3.keyid"] if "s3.keyid" in __opts__ else None
    service_url = __opts__["s3.service_url"] if "s3.service_url" in __opts__ else None
    verify_ssl = __opts__["s3.verify_ssl"] if "s3.verify_ssl" in __opts__ else None
    kms_keyid = __opts__["aws.kmw.keyid"] if "aws.kms.keyid" in __opts__ else None
    location = __opts__["s3.location"] if "s3.location" in __opts__ else None
    path_style = __opts__["s3.path_style"] if "s3.path_style" in __opts__ else None
    https_enable = (
        __opts__["s3.https_enable"] if "s3.https_enable" in __opts__ else None
    )

    return (
        key,
        keyid,
        service_url,
        verify_ssl,
        kms_keyid,
        location,
        path_style,
        https_enable,
    )


def _init():
    """
    Connect to S3 and download the metadata for each file in all buckets
    specified and cache the data to disk.
    """
    cache_file = _get_buckets_cache_filename()
    exp = time.time() - S3_CACHE_EXPIRE

    # check mtime of the buckets files cache
    metadata = None
    try:
        if os.path.getmtime(cache_file) > exp:
            metadata = _read_buckets_cache_file(cache_file)
    except OSError:
        pass

    if metadata is None:
        # bucket files cache expired or does not exist
        metadata = _refresh_buckets_cache_file(cache_file)

    return metadata


def _get_cache_dir():
    """
    Return the path to the s3cache dir
    """

    # Or is that making too many assumptions?
    return os.path.join(__opts__["cachedir"], "s3cache")


def _get_cached_file_name(bucket_name, saltenv, path):
    """
    Return the cached file name for a bucket path file
    """

    file_path = os.path.join(_get_cache_dir(), saltenv, bucket_name, path)

    # make sure bucket and saltenv directories exist
    if not os.path.exists(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))

    return file_path


def _get_buckets_cache_filename():
    """
    Return the filename of the cache for bucket contents.
    Create the path if it does not exist.
    """

    cache_dir = _get_cache_dir()
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    return os.path.join(cache_dir, "buckets_files.cache")


def _refresh_buckets_cache_file(cache_file):
    """
    Retrieve the content of all buckets and cache the metadata to the buckets
    cache file
    """

    log.debug("Refreshing buckets cache file")

    (
        key,
        keyid,
        service_url,
        verify_ssl,
        kms_keyid,
        location,
        path_style,
        https_enable,
    ) = _get_s3_key()
    metadata = {}

    # helper s3 query function
    def __get_s3_meta(bucket, key=key, keyid=keyid):
        ret, marker = [], ""
        while True:
            tmp = __utils__["s3.query"](
                key=key,
                keyid=keyid,
                kms_keyid=keyid,
                bucket=bucket,
                service_url=service_url,
                verify_ssl=verify_ssl,
                location=location,
                return_bin=False,
                path_style=path_style,
                https_enable=https_enable,
                params={"marker": marker},
            )
            headers = []
            for header in tmp:
                if "Key" in header:
                    break
                headers.append(header)
            ret.extend(tmp)
            if all(
                [header.get("IsTruncated", "false") == "false" for header in headers]
            ):
                break
            marker = tmp[-1]["Key"]
        return ret

    if _is_env_per_bucket():
        # Single environment per bucket
        for saltenv, buckets in _get_buckets().items():
            bucket_files_list = []
            for bucket_name in buckets:
                bucket_files = {}
                s3_meta = __get_s3_meta(bucket_name)

                # s3 query returned nothing
                if not s3_meta:
                    continue

                # grab only the files/dirs
                bucket_files[bucket_name] = [k for k in s3_meta if "Key" in k]
                bucket_files_list.append(bucket_files)

                # check to see if we added any keys, otherwise investigate possible error conditions
                if not bucket_files[bucket_name]:
                    meta_response = {}
                    for k in s3_meta:
                        if "Code" in k or "Message" in k:
                            # assumes no duplicate keys, consisdent with current erro response.
                            meta_response.update(k)
                    # attempt use of human readable output first.
                    try:
                        log.warning(
                            "'%s' response for bucket '%s'",
                            meta_response["Message"],
                            bucket_name,
                        )
                        continue
                    except KeyError:
                        # no human readable error message provided
                        if "Code" in meta_response:
                            log.warning(
                                "'%s' response for bucket '%s'",
                                meta_response["Code"],
                                bucket_name,
                            )
                            continue
                        else:
                            log.warning(
                                "S3 Error! Do you have any files in your S3 bucket?"
                            )
                            return {}

            metadata[saltenv] = bucket_files_list

    else:
        # Multiple environments per buckets
        for bucket_name in _get_buckets():
            s3_meta = __get_s3_meta(bucket_name)

            # s3 query returned nothing
            if not s3_meta:
                continue

            # pull out the environment dirs (e.g. the root dirs)
            files = [k for k in s3_meta if "Key" in k]

            # check to see if we added any keys, otherwise investigate possible error conditions
            if not files:
                meta_response = {}
                for k in s3_meta:
                    if "Code" in k or "Message" in k:
                        # assumes no duplicate keys, consisdent with current erro response.
                        meta_response.update(k)
                # attempt use of human readable output first.
                try:
                    log.warning(
                        "'%s' response for bucket '%s'",
                        meta_response["Message"],
                        bucket_name,
                    )
                    continue
                except KeyError:
                    # no human readable error message provided
                    if "Code" in meta_response:
                        log.warning(
                            "'%s' response for bucket '%s'",
                            meta_response["Code"],
                            bucket_name,
                        )
                        continue
                    else:
                        log.warning(
                            "S3 Error! Do you have any files in your S3 bucket?"
                        )
                        return {}

            environments = [(os.path.dirname(k["Key"]).split("/", 1))[0] for k in files]
            environments = set(environments)

            # pull out the files for the environment
            for saltenv in environments:
                # grab only files/dirs that match this saltenv
                env_files = [k for k in files if k["Key"].startswith(saltenv)]

                if saltenv not in metadata:
                    metadata[saltenv] = []

                found = False
                for bucket_files in metadata[saltenv]:
                    if bucket_name in bucket_files:
                        bucket_files[bucket_name] += env_files
                        found = True
                        break
                if not found:
                    metadata[saltenv].append({bucket_name: env_files})

    # write the metadata to disk
    _write_buckets_cache_file(metadata, cache_file)

    return metadata


def _write_buckets_cache_file(metadata, cache_file):
    """
    Write the contents of the buckets cache file
    """
    if os.path.isfile(cache_file):
        os.remove(cache_file)

    log.debug("Writing buckets cache file")

    with salt.utils.files.fopen(cache_file, "wb") as fp_:
        pickle.dump(metadata, fp_)


def _read_buckets_cache_file(cache_file):
    """
    Return the contents of the buckets cache file
    """

    log.debug("Reading buckets cache file")

    with salt.utils.files.fopen(cache_file, "rb") as fp_:
        try:
            data = pickle.load(fp_)
        except (
            pickle.UnpicklingError,
            AttributeError,
            EOFError,
            ImportError,
            IndexError,
            KeyError,
            ValueError,
        ) as exc:
            log.debug("Exception reading buckets cache file: '%s'", exc)
            data = None

    return data


def _find_files(metadata):
    """
    Looks for all the files in the S3 bucket cache metadata
    """

    ret = []
    found = {}

    for bucket_dict in metadata:
        for bucket_name, data in bucket_dict.items():
            filepaths = [k["Key"] for k in data]
            filepaths = [k for k in filepaths if not k.endswith("/")]
            if bucket_name not in found:
                found[bucket_name] = True
                ret.append({bucket_name: filepaths})
            else:
                for bucket in ret:
                    if bucket_name in bucket:
                        bucket[bucket_name] += filepaths
                        break
    return ret


def _find_dirs(metadata):
    """
    Looks for all the directories in the S3 bucket cache metadata.

    Supports trailing '/' keys (as created by S3 console) as well as
    directories discovered in the path of file keys.
    """

    ret = []
    found = {}

    for bucket_dict in metadata:
        for bucket_name, data in bucket_dict.items():
            dirpaths = set()
            for path in [k["Key"] for k in data]:
                prefix = ""
                for part in path.split("/")[:-1]:
                    directory = prefix + part + "/"
                    dirpaths.add(directory)
                    prefix = directory
            if bucket_name not in found:
                found[bucket_name] = True
                ret.append({bucket_name: list(dirpaths)})
            else:
                for bucket in ret:
                    if bucket_name in bucket:
                        bucket[bucket_name] += list(dirpaths)
                        bucket[bucket_name] = list(set(bucket[bucket_name]))
                        break
    return ret


def _find_file_meta(metadata, bucket_name, saltenv, path):
    """
    Looks for a file's metadata in the S3 bucket cache file
    """
    env_meta = metadata[saltenv] if saltenv in metadata else {}
    bucket_meta = {}
    for bucket in env_meta:
        if bucket_name in bucket:
            bucket_meta = bucket[bucket_name]
    files_meta = list(list(filter((lambda k: "Key" in k), bucket_meta)))

    for item_meta in files_meta:
        if "Key" in item_meta and item_meta["Key"] == path:
            try:
                # Get rid of quotes surrounding md5
                item_meta["ETag"] = item_meta["ETag"].strip('"')
            except KeyError:
                pass
            return item_meta


def _get_buckets():
    """
    Return the configuration buckets
    """

    return __opts__["s3.buckets"] if "s3.buckets" in __opts__ else {}


def _get_file_from_s3(metadata, saltenv, bucket_name, path, cached_file_path):
    """
    Checks the local cache for the file, if it's old or missing go grab the
    file from S3 and update the cache
    """
    (
        key,
        keyid,
        service_url,
        verify_ssl,
        kms_keyid,
        location,
        path_style,
        https_enable,
    ) = _get_s3_key()

    # check the local cache...
    if os.path.isfile(cached_file_path):
        file_meta = _find_file_meta(metadata, bucket_name, saltenv, path)
        if file_meta:
            file_etag = file_meta["ETag"]

            if file_etag.find("-") == -1:
                file_md5 = file_etag
                cached_md5 = salt.utils.hashutils.get_hash(cached_file_path, "md5")

                # hashes match we have a cache hit
                if cached_md5 == file_md5:
                    return
            else:
                cached_file_stat = os.stat(cached_file_path)
                cached_file_size = cached_file_stat.st_size
                cached_file_mtime = datetime.datetime.fromtimestamp(
                    cached_file_stat.st_mtime
                )

                cached_file_lastmod = datetime.datetime.strptime(
                    file_meta["LastModified"], "%Y-%m-%dT%H:%M:%S.%fZ"
                )
                if (
                    cached_file_size == int(file_meta["Size"])
                    and cached_file_mtime > cached_file_lastmod
                ):
                    log.debug(
                        "cached file size equal to metadata size and "
                        "cached file mtime later than metadata last "
                        "modification time."
                    )
                    ret = __utils__["s3.query"](
                        key=key,
                        keyid=keyid,
                        kms_keyid=keyid,
                        method="HEAD",
                        bucket=bucket_name,
                        service_url=service_url,
                        verify_ssl=verify_ssl,
                        location=location,
                        path=urllib.parse.quote(path),
                        local_file=cached_file_path,
                        full_headers=True,
                        path_style=path_style,
                        https_enable=https_enable,
                    )
                    if ret is not None:
                        for header_name, header_value in ret["headers"].items():
                            name = header_name.strip()
                            value = header_value.strip()
                            if str(name).lower() == "last-modified":
                                s3_file_mtime = datetime.datetime.strptime(
                                    value, "%a, %d %b %Y %H:%M:%S %Z"
                                )
                            elif str(name).lower() == "content-length":
                                s3_file_size = int(value)
                        if (
                            cached_file_size == s3_file_size
                            and cached_file_mtime > s3_file_mtime
                        ):
                            log.info(
                                "%s - %s : %s skipped download since cached file size "
                                "equal to and mtime after s3 values",
                                bucket_name,
                                saltenv,
                                path,
                            )
                            return

    # ... or get the file from S3
    __utils__["s3.query"](
        key=key,
        keyid=keyid,
        kms_keyid=keyid,
        bucket=bucket_name,
        service_url=service_url,
        verify_ssl=verify_ssl,
        location=location,
        path=urllib.parse.quote(path),
        local_file=cached_file_path,
        path_style=path_style,
        https_enable=https_enable,
    )


def _trim_env_off_path(paths, saltenv, trim_slash=False):
    """
    Return a list of file paths with the saltenv directory removed
    """
    env_len = None if _is_env_per_bucket() else len(saltenv) + 1
    slash_len = -1 if trim_slash else None

    return [d[env_len:slash_len] for d in paths]


def _is_env_per_bucket():
    """
    Return the configuration mode, either buckets per environment or a list of
    buckets that have environment dirs in their root
    """

    buckets = _get_buckets()
    if isinstance(buckets, dict):
        return True
    elif isinstance(buckets, list):
        return False
    else:
        raise ValueError("Incorrect s3.buckets type given in config")



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import salt.modules.mac_sysctl as mac_sysctl
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import DEFAULT, MagicMock, mock_open, patch
from tests.support.unit import TestCase


class DarwinSysctlTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.mac_sysctl module
    """

    def setup_loader_modules(self):
        return {mac_sysctl: {}}

    def test_get(self):
        """
        Tests the return of get function
        """
        mock_cmd = MagicMock(return_value="foo")
        with patch.dict(mac_sysctl.__salt__, {"cmd.run": mock_cmd}):
            self.assertEqual(mac_sysctl.get("kern.ostype"), "foo")

    def test_assign_cmd_failed(self):
        """
        Tests if the assignment was successful or not
        """
        cmd = {
            "pid": 3548,
            "retcode": 1,
            "stderr": "",
            "stdout": "net.inet.icmp.icmplim: 250 -> 50",
        }
        mock_cmd = MagicMock(return_value=cmd)
        with patch.dict(mac_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
            self.assertRaises(
                CommandExecutionError, mac_sysctl.assign, "net.inet.icmp.icmplim", 50
            )

    def test_assign(self):
        """
        Tests the return of successful assign function
        """
        cmd = {
            "pid": 3548,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.inet.icmp.icmplim: 250 -> 50",
        }
        ret = {"net.inet.icmp.icmplim": "50"}
        mock_cmd = MagicMock(return_value=cmd)
        with patch.dict(mac_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
            self.assertEqual(mac_sysctl.assign("net.inet.icmp.icmplim", 50), ret)

    def test_persist_no_conf_failure(self):
        """
        Tests adding of config file failure
        """
        read_data = IOError(13, "Permission denied", "/file")
        with patch("salt.utils.files.fopen", mock_open(read_data=read_data)), patch(
            "os.path.isfile", MagicMock(return_value=False)
        ):
            self.assertRaises(
                CommandExecutionError,
                mac_sysctl.persist,
                "net.inet.icmp.icmplim",
                50,
                config=None,
            )

    def test_persist_no_conf_success(self):
        """
        Tests successful add of config file when it did not already exist
        """
        config = "/etc/sysctl.conf"
        isfile_mock = MagicMock(side_effect=lambda x: False if x == config else DEFAULT)
        with patch("salt.utils.files.fopen", mock_open()) as m_open, patch(
            "os.path.isfile", isfile_mock
        ):
            mac_sysctl.persist("net.inet.icmp.icmplim", 50, config=config)
            # We only should have opened the one file
            num_handles = len(m_open.filehandles)
            assert num_handles == 1, num_handles
            writes = m_open.write_calls()
            # We should have called .write() only once, with the expected
            # content
            num_writes = len(writes)
            assert num_writes == 1, num_writes
            assert writes[0] == "#\n# Kernel sysctl configuration\n#\n", writes[0]

    def test_persist_success(self):
        """
        Tests successful write to existing sysctl file
        """
        config = "/etc/sysctl.conf"
        to_write = "#\n# Kernel sysctl configuration\n#\n"
        writelines_calls = [
            [
                "#\n",
                "# Kernel sysctl configuration\n",
                "#\n",
                "net.inet.icmp.icmplim=50\n",
            ]
        ]
        isfile_mock = MagicMock(side_effect=lambda x: True if x == config else DEFAULT)
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=to_write)
        ) as m_open, patch("os.path.isfile", isfile_mock):
            mac_sysctl.persist("net.inet.icmp.icmplim", 50, config=config)
            # We only should have opened the one file
            num_handles = len(m_open.filehandles)
            assert num_handles == 1, num_handles
            writes = m_open.writelines_calls()
            assert writes == writelines_calls, writes



"""
Loader mechanism for caching data, with data expiration, etc.

.. versionadded:: 2016.11.0
"""


import logging
import time

import salt.config
import salt.loader
import salt.syspaths
from salt.utils.odict import OrderedDict

log = logging.getLogger(__name__)


def factory(opts, **kwargs):
    """
    Creates and returns the cache class.
    If memory caching is enabled by opts MemCache class will be instantiated.
    If not Cache class will be returned.
    """
    if opts.get("memcache_expire_seconds", 0):
        cls = MemCache
    else:
        cls = Cache
    return cls(opts, **kwargs)


class Cache:
    """
    Base caching object providing access to the modular cache subsystem.

    Related configuration options:

    :param cache:
        The name of the cache driver to use. This is the name of the python
        module of the `salt.cache` package. Default is `localfs`.

    Terminology.

    Salt cache subsystem is organized as a tree with nodes and leafs like a
    filesystem. Cache consists of banks. Each bank can contain a number of
    keys. Each key can contain a dict or any other object serializable with
    `salt.payload`. I.e. any data object in the cache can be
    addressed by the path to the bank and the key name:
        bank: 'minions/alpha'
        key:  'data'

    Bank names should be formatted in a way that can be used as a
    directory structure. If slashes are included in the name, then they
    refer to a nested structure.

    Key name is a string identifier of a data container (like a file inside a
    directory) which will hold the data.
    """

    def __init__(self, opts, cachedir=None, **kwargs):
        self.opts = opts
        if cachedir is None:
            self.cachedir = opts.get("cachedir", salt.syspaths.CACHE_DIR)
        else:
            self.cachedir = cachedir
        self.driver = opts.get("cache", salt.config.DEFAULT_MASTER_OPTS["cache"])
        self._modules = None
        self._kwargs = kwargs
        self._kwargs["cachedir"] = self.cachedir

    def __lazy_init(self):
        self._modules = salt.loader.cache(self.opts)
        fun = "{}.init_kwargs".format(self.driver)
        if fun in self.modules:
            self._kwargs = self.modules[fun](self._kwargs)
        else:
            self._kwargs = {}

    @property
    def modules(self):
        if self._modules is None:
            self.__lazy_init()
        return self._modules

    def cache(self, bank, key, fun, loop_fun=None, **kwargs):
        """
        Check cache for the data. If it is there, check to see if it needs to
        be refreshed.

        If the data is not there, or it needs to be refreshed, then call the
        callback function (``fun``) with any given ``**kwargs``.

        In some cases, the callback function returns a list of objects which
        need to be processed by a second function. If that is the case, then
        the second function is passed in as ``loop_fun``. Each item in the
        return list from the first function will be the only argument for the
        second function.
        """
        expire_seconds = kwargs.get("expire", 86400)  # 1 day

        updated = self.updated(bank, key)
        update_cache = False
        if updated is None:
            update_cache = True
        else:
            if int(time.time()) - updated > expire_seconds:
                update_cache = True

        data = self.fetch(bank, key)

        if not data or update_cache is True:
            if loop_fun is not None:
                data = []
                items = fun(**kwargs)
                for item in items:
                    data.append(loop_fun(item))
            else:
                data = fun(**kwargs)
            self.store(bank, key, data)

        return data

    def store(self, bank, key, data):
        """
        Store data using the specified module

        :param bank:
            The name of the location inside the cache which will hold the key
            and its associated data.

        :param key:
            The name of the key (or file inside a directory) which will hold
            the data. File extensions should not be provided, as they will be
            added by the driver itself.

        :param data:
            The data which will be stored in the cache. This data should be
            in a format which can be serialized by msgpack.

        :raises SaltCacheError:
            Raises an exception if cache driver detected an error accessing data
            in the cache backend (auth, permissions, etc).
        """
        fun = "{}.store".format(self.driver)
        return self.modules[fun](bank, key, data, **self._kwargs)

    def fetch(self, bank, key):
        """
        Fetch data using the specified module

        :param bank:
            The name of the location inside the cache which will hold the key
            and its associated data.

        :param key:
            The name of the key (or file inside a directory) which will hold
            the data. File extensions should not be provided, as they will be
            added by the driver itself.

        :return:
            Return a python object fetched from the cache or an empty dict if
            the given path or key not found.

        :raises SaltCacheError:
            Raises an exception if cache driver detected an error accessing data
            in the cache backend (auth, permissions, etc).
        """
        fun = "{}.fetch".format(self.driver)
        return self.modules[fun](bank, key, **self._kwargs)

    def updated(self, bank, key):
        """
        Get the last updated epoch for the specified key

        :param bank:
            The name of the location inside the cache which will hold the key
            and its associated data.

        :param key:
            The name of the key (or file inside a directory) which will hold
            the data. File extensions should not be provided, as they will be
            added by the driver itself.

        :return:
            Return an int epoch time in seconds or None if the object wasn't
            found in cache.

        :raises SaltCacheError:
            Raises an exception if cache driver detected an error accessing data
            in the cache backend (auth, permissions, etc).
        """
        fun = "{}.updated".format(self.driver)
        return self.modules[fun](bank, key, **self._kwargs)

    def flush(self, bank, key=None):
        """
        Remove the key from the cache bank with all the key content. If no key is specified remove
        the entire bank with all keys and sub-banks inside.

        :param bank:
            The name of the location inside the cache which will hold the key
            and its associated data.

        :param key:
            The name of the key (or file inside a directory) which will hold
            the data. File extensions should not be provided, as they will be
            added by the driver itself.

        :raises SaltCacheError:
            Raises an exception if cache driver detected an error accessing data
            in the cache backend (auth, permissions, etc).
        """
        fun = "{}.flush".format(self.driver)
        return self.modules[fun](bank, key=key, **self._kwargs)

    def list(self, bank):
        """
        Lists entries stored in the specified bank.

        :param bank:
            The name of the location inside the cache which will hold the key
            and its associated data.

        :return:
            An iterable object containing all bank entries. Returns an empty
            iterator if the bank doesn't exists.

        :raises SaltCacheError:
            Raises an exception if cache driver detected an error accessing data
            in the cache backend (auth, permissions, etc).
        """
        fun = "{}.list".format(self.driver)
        return self.modules[fun](bank, **self._kwargs)

    def contains(self, bank, key=None):
        """
        Checks if the specified bank contains the specified key.

        :param bank:
            The name of the location inside the cache which will hold the key
            and its associated data.

        :param key:
            The name of the key (or file inside a directory) which will hold
            the data. File extensions should not be provided, as they will be
            added by the driver itself.

        :return:
            Returns True if the specified key exists in the given bank and False
            if not.
            If key is None checks for the bank existense.

        :raises SaltCacheError:
            Raises an exception if cache driver detected an error accessing data
            in the cache backend (auth, permissions, etc).
        """
        fun = "{}.contains".format(self.driver)
        return self.modules[fun](bank, key, **self._kwargs)


class MemCache(Cache):
    """
    Short-lived in-memory cache store keeping values on time and/or size (count)
    basis.
    """

    # {<storage_id>: odict({<key>: [atime, data], ...}), ...}
    data = {}

    def __init__(self, opts, **kwargs):
        super().__init__(opts, **kwargs)
        self.expire = opts.get("memcache_expire_seconds", 10)
        self.max = opts.get("memcache_max_items", 1024)
        self.cleanup = opts.get("memcache_full_cleanup", False)
        self.debug = opts.get("memcache_debug", False)
        if self.debug:
            self.call = 0
            self.hit = 0
        self._storage = None

    @classmethod
    def __cleanup(cls, expire):
        now = time.time()
        for storage in cls.data.values():
            for key, data in list(storage.items()):
                if data[0] + expire < now:
                    del storage[key]
                else:
                    break

    def _get_storage_id(self):
        fun = "{}.storage_id".format(self.driver)
        if fun in self.modules:
            return self.modules[fun](self.kwargs)
        else:
            return self.driver

    @property
    def storage(self):
        if self._storage is None:
            storage_id = self._get_storage_id()
            if storage_id not in MemCache.data:
                MemCache.data[storage_id] = OrderedDict()
            self._storage = MemCache.data[storage_id]
        return self._storage

    def fetch(self, bank, key):
        if self.debug:
            self.call += 1
        now = time.time()
        record = self.storage.pop((bank, key), None)
        # Have a cached value for the key
        if record is not None and record[0] + self.expire >= now:
            if self.debug:
                self.hit += 1
                log.debug(
                    "MemCache stats (call/hit/rate): %s/%s/%s",
                    self.call,
                    self.hit,
                    float(self.hit) / self.call,
                )
            # update atime and return
            record[0] = now
            self.storage[(bank, key)] = record
            return record[1]

        # Have no value for the key or value is expired
        data = super().fetch(bank, key)
        if len(self.storage) >= self.max:
            if self.cleanup:
                MemCache.__cleanup(self.expire)
            if len(self.storage) >= self.max:
                self.storage.popitem(last=False)
        self.storage[(bank, key)] = [now, data]
        return data

    def store(self, bank, key, data):
        self.storage.pop((bank, key), None)
        super().store(bank, key, data)
        if len(self.storage) >= self.max:
            if self.cleanup:
                MemCache.__cleanup(self.expire)
            if len(self.storage) >= self.max:
                self.storage.popitem(last=False)
        self.storage[(bank, key)] = [time.time(), data]

    def flush(self, bank, key=None):
        self.storage.pop((bank, key), None)
        super().flush(bank, key)



from salt.utils import dictdiffer
from salt.utils.listdiffer import list_diff
from tests.support.unit import TestCase

NONE = dictdiffer.RecursiveDictDiffer.NONE_VALUE


class ListDictDifferTestCase(TestCase):
    def setUp(self):
        old_list = [
            {"key": 1, "value": "foo1", "int_value": 101},
            {"key": 2, "value": "foo2", "int_value": 102},
            {"key": 3, "value": "foo3", "int_value": 103},
        ]
        new_list = [
            {"key": 1, "value": "foo1", "int_value": 101},
            {"key": 2, "value": "foo2", "int_value": 112},
            {"key": 5, "value": "foo5", "int_value": 105},
        ]
        self.list_diff = list_diff(old_list, new_list, key="key")

    def tearDown(self):
        for attrname in ("list_diff",):
            try:
                delattr(self, attrname)
            except AttributeError:
                continue

    def test_added(self):
        self.assertEqual(len(self.list_diff.added), 1)
        self.assertDictEqual(
            self.list_diff.added[0], {"key": 5, "value": "foo5", "int_value": 105}
        )

    def test_removed(self):
        self.assertEqual(len(self.list_diff.removed), 1)
        self.assertDictEqual(
            self.list_diff.removed[0], {"key": 3, "value": "foo3", "int_value": 103}
        )

    def test_diffs(self):
        self.assertEqual(len(self.list_diff.diffs), 3)
        self.assertDictEqual(
            self.list_diff.diffs[0], {2: {"int_value": {"new": 112, "old": 102}}}
        )
        self.assertDictEqual(
            self.list_diff.diffs[1],
            # Added items
            {
                5: {
                    "int_value": {"new": 105, "old": NONE},
                    "key": {"new": 5, "old": NONE},
                    "value": {"new": "foo5", "old": NONE},
                }
            },
        )
        self.assertDictEqual(
            self.list_diff.diffs[2],
            # Removed items
            {
                3: {
                    "int_value": {"new": NONE, "old": 103},
                    "key": {"new": NONE, "old": 3},
                    "value": {"new": NONE, "old": "foo3"},
                }
            },
        )

    def test_new_values(self):
        self.assertEqual(len(self.list_diff.new_values), 2)
        self.assertDictEqual(self.list_diff.new_values[0], {"key": 2, "int_value": 112})
        self.assertDictEqual(
            self.list_diff.new_values[1], {"key": 5, "value": "foo5", "int_value": 105}
        )

    def test_old_values(self):
        self.assertEqual(len(self.list_diff.old_values), 2)
        self.assertDictEqual(self.list_diff.old_values[0], {"key": 2, "int_value": 102})
        self.assertDictEqual(
            self.list_diff.old_values[1], {"key": 3, "value": "foo3", "int_value": 103}
        )

    def test_changed_all(self):
        self.assertEqual(
            self.list_diff.changed(selection="all"),
            [
                "key.2.int_value",
                "key.5.int_value",
                "key.5.value",
                "key.3.int_value",
                "key.3.value",
            ],
        )

    def test_changed_intersect(self):
        self.assertEqual(
            self.list_diff.changed(selection="intersect"), ["key.2.int_value"]
        )

    def test_changes_str(self):
        self.assertEqual(
            self.list_diff.changes_str,
            "\tidentified by key 2:\n"
            "\tint_value from 102 to 112\n"
            "\tidentified by key 3:\n"
            "\twill be removed\n"
            "\tidentified by key 5:\n"
            "\twill be added\n",
        )



import pytest
from tests.support.case import ModuleCase


@pytest.mark.flaky(max_runs=4)
@pytest.mark.skip_unless_on_windows
@pytest.mark.windows_whitelisted
class NTPTest(ModuleCase):
    """
    Validate windows ntp module
    """

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_ntp_set_servers(self):
        """
        test ntp get and set servers
        """
        ntp_srv = "pool.ntp.org"
        set_srv = self.run_function("ntp.set_servers", [ntp_srv])
        self.assertTrue(set_srv)

        get_srv = self.run_function("ntp.get_servers")
        self.assertEqual(ntp_srv, get_srv[0])
